import wx

if 'phoenix' in wx.version():
    bsm_is_phoenix = True
    import wx.html
    BitmapFromXPMData = wx.Bitmap
    HtmlListBox = wx.html.HtmlListBox
    ArtProvider_GetBitmap = wx.ArtProvider.GetBitmap
    EmptyIcon = wx.Icon
    EmptyBitmap = wx.Bitmap
    EVT_COMMAND_FIND = wx.EVT_FIND
    EVT_COMMAND_FIND_NEXT = wx.EVT_FIND_NEXT
    EVT_COMMAND_FIND_REPLACE = wx.EVT_FIND_REPLACE
    EVT_COMMAND_FIND_REPLACE_ALL = wx.EVT_FIND_REPLACE_ALL
    EVT_COMMAND_FIND_CLOSE = wx.EVT_FIND_CLOSE
    SystemSettings_GetColour = wx.SystemSettings.GetColour
    NamedColour = wx.Colour
    StockCursor = wx.Cursor
    PyDropTarget = wx.DropTarget
    PyTextDataObject = wx.TextDataObject
    APP_ASSERT_DIALOG = wx.APP_ASSERT_DIALOG
    FD_OPEN = wx.FD_OPEN
    FD_FILE_MUST_EXIST = wx.FD_FILE_MUST_EXIST
else:
    bsm_is_phoenix = False
    BitmapFromXPMData = wx.BitmapFromXPMData
    HtmlListBox = wx.HtmlListBox
    ArtProvider_GetBitmap = wx.ArtProvider_GetBitmap
    EmptyIcon = wx.EmptyIcon
    EmptyBitmap = wx.EmptyBitmap
    EVT_COMMAND_FIND = wx.EVT_COMMAND_FIND
    EVT_COMMAND_FIND_NEXT = wx.EVT_COMMAND_FIND_NEXT
    EVT_COMMAND_FIND_REPLACE = wx.EVT_COMMAND_FIND_REPLACE
    EVT_COMMAND_FIND_REPLACE_ALL = wx.EVT_COMMAND_FIND_REPLACE_ALL
    EVT_COMMAND_FIND_CLOSE = wx.EVT_COMMAND_FIND_CLOSE
    SystemSettings_GetColour = wx.SystemSettings_GetColour
    NamedColour = wx.NamedColour
    StockCursor = wx.StockCursor
    PyDropTarget = wx.PyDropTarget
    PyTextDataObject = wx.PyTextDataObject
    APP_ASSERT_DIALOG = wx.PYAPP_ASSERT_DIALOG
    FD_OPEN = wx.OPEN
    FD_FILE_MUST_EXIST = wx.FILE_MUST_EXIST


def SetClippingRect(dc, rc):
    if bsm_is_phoenix:
        dc.SetClippingRegion(rc)
    else:
        dc.SetClippingRect(rc)


def BitmapFromXPM(xpm):
    xpm_b = [x.encode('utf-8') for x in xpm]
    return BitmapFromXPMData(xpm_b)


def tbAddTool(tb, *args, **kwargs):
    if bsm_is_phoenix:
        tb.AddTool(*args, **kwargs)
    else:
        tb.AddLabelTool(*args, **kwargs)


def tbAddCheckTool(tb, *args, **kwargs):
    if bsm_is_phoenix:
        tb.AddCheckTool(*args, **kwargs)
    else:
        tb.AddCheckLabelTool(*args, **kwargs)


def treeGetData(tree, *args, **kwargs):
    if bsm_is_phoenix:
        return tree.GetItemData(*args, **kwargs)
    else:
        data = tree.GetPyData(*args, **kwargs)
        return data


def treeAppendItem(tree, parent, label, img, selimg, data):
    if not bsm_is_phoenix:
        data = wx.TreeItemData(data)
    return tree.AppendItem(parent, label, img, selimg, data)


def menuAppend(menu, item):
    if not bsm_is_phoenix:
        return menu.AppendItem(item)
    else:
        return menu.Append(item)



"""
Manage windows (originated from matplotlib).
"""
import gc
import six


class Gcm(object):
    """
    It consists of two class
    attributes (a list and a dictionary), and a set of static
    methods that operate on those attributes, accessing them
    directly as class attributes.

    Attributes:

        *mgrs*:
          dictionary of the form {*num*: *manager*, ...}

        *_activeQue*:
          list of *managers*, with active one at the end

    """
    def __init__(self):
        self._activeQue = []
        self.mgrs = {}

    def get_manager(self, num):
        """
        If manager *num* exists, make it active
        and return the manager; otherwise return *None*.
        """
        manager = self.mgrs.get(num, None)
        if manager is not None:
            self.set_active(manager)
        return manager

    def destroy(self, num):
        """
        Try to remove all traces of manager *num*.

        In the interactive backends, this is bound to the
        window "destroy" and "delete" events.
        """
        if not self.has_num(num):
            return
        manager = self.mgrs[num]

        # There must be a good reason for theg following careful
        # rebuilding of the activeQue; what is it?
        oldQue = self._activeQue[:]
        self._activeQue = []
        for f in oldQue:
            if f != manager:
                self._activeQue.append(f)

        del self.mgrs[num]
        #manager.destroy()
        gc.collect(1)

    def destroy_mgr(self, mgr):
        "*mgr* is a manager instance"
        num = None
        for manager in six.itervalues(self.mgrs):
            if manager == mgr:
                num = manager.num
                break
        if num is not None:
            self.destroy(num)

    def destroy_all(self):
        """remove all the managers"""
        self._activeQue = []
        self.mgrs.clear()
        gc.collect(1)

    def has_num(self, num):
        """
        Return *True* if manager *num* exists.
        """
        return num in self.mgrs

    def get_all_managers(self):
        """
        Return a list of managers.
        """
        return list(self.mgrs.values())

    def get_num_managers(self):
        """
        Return the number of manager being managed.
        """
        return len(self.mgrs)

    def get_active(self):
        """
        Return the manager of the active manager, or *None*.
        """
        if len(self._activeQue) == 0:
            return None
        else:
            return self._activeQue[-1]

    def set_active(self, mgr):
        """
        Make the mgr corresponding to *manager* the active one.
        """
        oldQue = self._activeQue[:]
        self._activeQue = []
        for m in oldQue:
            if m != mgr:
                self._activeQue.append(m)
        self._activeQue.append(mgr)
        self.mgrs[mgr.num] = mgr

    def get_nums(self):
        """return all the occupied nums"""
        return list(self.mgrs.keys())

    def get_next_num(self):
        """return the next available num"""
        allnums = self.get_nums()
        return max(allnums) + 1 if allnums else 1



import pydoc
import os
import time
import traceback
import sys
import six
import wx
import wx.lib.agw.aui as aui
import wx.py.dispatcher as dp
import wx.html2 as html
import wx.svg
from .dirtreectrl import DirTreeCtrl, Directory
from .bsmxpm import backward_svg, forward_svg, goup_xpm, home_xpm
from .autocomplete import AutocompleteTextCtrl
from .utility import FastLoadTreeCtrl, svg_to_bitmap

from .. import to_byte

html_template = '''
<html>
    <head>
        <title>%(title)s</title>
    </head>
    <body>
    <FONT FACE= "Courier New">
    <pre>
%(message)s
    </pre>
    </body>
</html>
'''


class HelpPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent)

        self.html = html.WebView.New(self)

        agwStyle = aui.AUI_TB_OVERFLOW | aui.AUI_TB_PLAIN_BACKGROUND
        self.tb = aui.AuiToolBar(self, agwStyle=agwStyle)
        self.tb.AddSimpleTool(wx.ID_BACKWARD, 'Back',
                              svg_to_bitmap(backward_svg),
                              'Go the previous page')

        self.tb.AddSimpleTool(wx.ID_FORWARD, 'Forward',
                              svg_to_bitmap(forward_svg),
                              'Go to the next page')
        self.search = AutocompleteTextCtrl(self.tb, completer=self.completer)
        item = self.tb.AddControl(self.search)
        item.SetProportion(1)
        self.tb.Realize()

        # Setup the layout
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.tb, 0, wx.ALL | wx.EXPAND, 0)
        sizer.Add(self.html, 1, wx.ALL | wx.EXPAND, 0)
        self.SetSizer(sizer)

        self.history = []
        self.history_index = -1
        self.findStr = ""
        self.findFlags = html.WEBVIEW_FIND_DEFAULT | html.WEBVIEW_FIND_WRAP

        self.Bind(wx.EVT_TEXT_ENTER, self.OnDoSearch, self.search)
        self.Bind(html.EVT_WEBVIEW_NAVIGATING, self.OnWebViewNavigating,
                  self.html)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateUI)
        self.Bind(wx.EVT_TOOL, self.OnBack, id=wx.ID_BACKWARD)
        self.Bind(wx.EVT_TOOL, self.OnForward, id=wx.ID_FORWARD)
        self.Bind(wx.EVT_TOOL, self.OnShowFind, id=wx.ID_FIND)

        accel = [(wx.ACCEL_CTRL, ord('F'), wx.ID_FIND)]
        self.accel = wx.AcceleratorTable(accel)
        self.SetAcceleratorTable(self.accel)

    def OnShowFind(self, evt):
        self.html.Find("")
        findData = wx.FindReplaceData()
        dlg = wx.FindReplaceDialog(self, findData, "Find")
        dlg.findData = findData
        dlg.Bind(wx.EVT_FIND, self.OnFind)
        dlg.Bind(wx.EVT_FIND_NEXT, self.OnFind)
        dlg.Show(True)

    def OnFind(self, evt):
        self.findStr = evt.GetFindString()
        flags = evt.GetFlags()
        self.findFlags = html.WEBVIEW_FIND_DEFAULT | html.WEBVIEW_FIND_WRAP
        if wx.FR_WHOLEWORD & flags:
            self.findFlags |= html.WEBVIEW_FIND_ENTIRE_WORD
        if wx.FR_MATCHCASE & flags:
            self.findFlags |= html.WEBVIEW_FIND_MATCH_CASE
        if not (wx.FR_DOWN & flags):
            self.findFlags |= html.WEBVIEW_FIND_BACKWARDS
        self.html.Find(self.findStr, self.findFlags)

    def completer(self, query):
        response = dp.send(signal='shell.auto_complete_list', command=query)
        if response:
            root = query[0:query.rfind('.') + 1]
            remain = query[query.rfind('.') + 1:]
            remain = remain.lower()
            objs = [o for o in response[0][1] if o.lower().startswith(remain)]
            return objs, objs, len(query) - len(root)
        return [], [], 0

    def add_history(self, command):
        if not self.history or self.history[-1] != command:
            self.history.append(command)
            self.history_index = -1

    def show_help(self, command, addhistory=True):
        try:
            strhelp = pydoc.plain(pydoc.render_doc(str(command), "Help on %s"))
            htmlpage = html_template % ({'title': '', 'message': strhelp})
            self.html.SetPage(htmlpage, '')
            # do not use SetValue since it will trigger the text update event, which
            # will popup the auto complete list window
            self.search.ChangeValue(command)
            if addhistory:
                self.add_history(command)
        except:
            traceback.print_exc(file=sys.stdout)

    def OnDoSearch(self, evt):
        command = self.search.GetValue()
        self.show_help(command)

    def OnWebViewNavigating(self, evt):
        # this event happens prior to trying to get a resource
        strURL = evt.GetURL()
        if strURL[:8] == 'bsmhelp:':
            # This is how you can cancel loading a page.
            evt.Veto()
            command = strURL[8:]
            self.show_help(command)

    def OnUpdateUI(self, event):
        idx = event.GetId()
        h_idx = -1
        h_len = len(self.history)
        if h_len > 0:
            h_idx = self.history_index % h_len
        if idx == wx.ID_FORWARD:
            event.Enable(0 <= h_idx < h_len - 1)
        elif idx == wx.ID_BACKWARD:
            event.Enable(h_idx > 0)

    def OnBack(self, event):
        # the button is only enabled when history_index>0
        self.history_index -= 1
        command = self.history[self.history_index]
        self.show_help(command, False)

    def OnForward(self, event):
        # the button is only enable when history_index hasn't reached the last
        # one
        self.history_index += 1
        command = self.history[self.history_index]
        self.show_help(command, False)


class HistoryPanel(wx.Panel):
    ID_EXECUTE = wx.NewId()
    TIME_STAMP_HEADER = "#bsm"

    def __init__(self, parent):
        wx.Panel.__init__(self, parent)
        style = wx.TR_DEFAULT_STYLE | wx.TR_HIDE_ROOT |wx.TR_MULTIPLE |\
                wx.TR_HAS_VARIABLE_ROW_HEIGHT
        # no need to sort the commands, as they are naturally sorted by
        # execution time
        self.tree = FastLoadTreeCtrl(self,
                                     getchildren=self.get_children,
                                     style=style,
                                     sort=False)
        self.history = {}
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.tree, 1, wx.ALL | wx.EXPAND, 0)
        self.SetSizer(sizer)
        dp.connect(receiver=self.AddHistory, signal='Shell.addHistory')
        self.root = self.tree.AddRoot('The Root Item')
        self.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.OnActivate, self.tree)
        self.Bind(wx.EVT_TREE_ITEM_RIGHT_CLICK, self.OnRightClick, self.tree)
        self.Bind(wx.EVT_MENU, self.OnProcessEvent, id=wx.ID_COPY)
        self.Bind(wx.EVT_MENU, self.OnProcessEvent, id=wx.ID_CUT)
        self.Bind(wx.EVT_MENU, self.OnProcessEvent, id=self.ID_EXECUTE)
        self.Bind(wx.EVT_MENU, self.OnProcessEvent, id=wx.ID_DELETE)
        self.Bind(wx.EVT_MENU, self.OnProcessEvent, id=wx.ID_CLEAR)

        accel = [
            (wx.ACCEL_CTRL, ord('C'), wx.ID_COPY),
            (wx.ACCEL_CTRL, ord('X'), wx.ID_CUT),
            (wx.ACCEL_CTRL, ord('E'), self.ID_EXECUTE),
            (wx.ACCEL_NORMAL, wx.WXK_DELETE, wx.ID_DELETE),
        ]
        self.accel = wx.AcceleratorTable(accel)
        self.SetAcceleratorTable(self.accel)
        self.LoadHistory()

    def Destroy(self):
        dp.disconnect(receiver=self.AddHistory, signal='Shell.addHistory')
        super(HistoryPanel, self).Destroy()

    def get_children(self, item):
        """ callback function to return the children of item """
        if item == self.tree.GetRootItem():
            childlist = list(six.iterkeys(self.history))
            # sort by time-stamp
            childlist.sort()
            is_folder = True
            clr = wx.Colour(100, 174, 100)
        else:
            stamp = self.tree.GetItemText(item)
            childlist = self.history.get(stamp, [])
            is_folder = False
            clr = None
            # free the list
            self.history.pop(stamp, None)
        children = []
        for obj in reversed(childlist):
            child = {
                'label': obj,
                'img': -1,
                'imgsel': -1,
                'data': '',
                'color': clr
            }
            child['is_folder'] = is_folder
            children.append(child)
        return children

    def LoadHistory(self):
        resp = dp.send('frame.get_config', group='shell', key='history')
        history = []
        if resp and resp[0][1]:
            history = resp[0][1]

        stamp = time.strftime('#%Y/%m/%d')
        for i in six.moves.range(len(history) - 1, -1, -1):
            value = history[i]
            if value.startswith(self.TIME_STAMP_HEADER):
                stamp = value[len(self.TIME_STAMP_HEADER):]
                self.history[stamp] = self.history.get(stamp, [])
            elif self.history.get(stamp, None) is not None:
                self.history[stamp].append(value)

        self.tree.FillChildren(self.root)
        item, cookie = self.tree.GetFirstChild(self.root)
        if item.IsOk():
            self.tree.Expand(item)
            child, _ = self.tree.GetFirstChild(item)
            if child.IsOk():
                self.tree.SelectItem(child)
                self.tree.EnsureVisible(child)

    def SaveHistory(self, config):
        """save the history"""
        config.DeleteGroup('/CommandHistory')
        config.SetPath('/CommandHistory')
        (item, cookie) = self.tree.GetFirstChild(self.root)
        pos = 0
        while item.IsOk():
            stamp = self.tree.GetItemText(item)
            config.Write("item%d" % pos, self.TIME_STAMP_HEADER + stamp)
            pos += 1

            childitem, childcookie = self.tree.GetFirstChild(item)
            if childitem.IsOk() and self.tree.GetItemText(childitem) != "...":
                while childitem.IsOk():
                    config.Write("item%d" % pos,
                                 self.tree.GetItemText(childitem))
                    childitem, childcookie = self.tree.GetNextChild(
                        item, childcookie)
                    pos = pos + 1
            # save the unexpanded folder
            for child in self.history.get(stamp, []):
                config.Write("item%d" % pos, child)
                pos = pos + 1

            (item, cookie) = self.tree.GetNextChild(self.root, cookie)

    def AddHistory(self, command, stamp=""):
        """ add history to treectrl """
        command = command.strip()
        if not command:
            return
        if not stamp:
            stamp = time.strftime('#%Y/%m/%d')

        # search the time stamp
        pos = 0
        item, cookie = self.tree.GetFirstChild(self.root)
        while item.IsOk():
            if self.tree.GetItemText(item) == stamp:
                break
            elif self.tree.GetItemText(item) > stamp:
                item = self.tree.InsertItemBefore(self.root, pos, stamp)
                self.tree.SetItemTextColour(item, wx.Colour(100, 174, 100))
                break
            pos = pos + 1
            (item, cookie) = self.tree.GetNextChild(self.root, cookie)
        # not find the time stamp, create one
        if not item.IsOk():
            item = self.tree.PrependItem(self.root, stamp)
            self.tree.SetItemTextColour(item, wx.Colour(100, 174, 100))
        # append the history
        if item.IsOk():
            self.tree.Expand(item)
            child = self.tree.PrependItem(item, command)
            self.tree.EnsureVisible(child)

    def OnActivate(self, event):
        item = event.GetItem()
        if not self.tree.ItemHasChildren(item):
            command = self.tree.GetItemText(item)
            dp.send(signal='shell.run', command=command)

    def OnRightClick(self, event):
        menu = wx.Menu()
        menu.Append(wx.ID_COPY, "Copy\tCtrl+C")
        menu.Append(wx.ID_CUT, "Cut\tCtrl+X")
        menu.Append(self.ID_EXECUTE, "Evaluate\tCtrl+E")
        menu.AppendSeparator()
        menu.AppendSeparator()
        menu.Append(wx.ID_DELETE, "Delete\tDel")
        menu.Append(wx.ID_CLEAR, "Clear history")
        self.PopupMenu(menu)
        menu.Destroy()

    def OnProcessEvent(self, event):
        items = self.tree.GetSelections()
        cmd = []
        for item in items:
            cmd.append(self.tree.GetItemText(item))
        evtId = event.GetId()
        if evtId in (wx.ID_COPY, wx.ID_CUT):
            clipData = wx.TextDataObject()
            clipData.SetText("\n".join(cmd))
            wx.TheClipboard.Open()
            wx.TheClipboard.SetData(clipData)
            wx.TheClipboard.Close()
            if evtId == wx.ID_CUT:
                for item in items:
                    if self.tree.ItemHasChildren(item):
                        self.tree.DeleteChildren(item)
                    self.tree.Delete(item)
        elif evtId == self.ID_EXECUTE:
            for c in cmd:
                dp.send(signal='shell.run', command=c)
        elif evtId == wx.ID_DELETE:
            for item in items:
                if self.tree.ItemHasChildren(item):
                    self.tree.DeleteChildren(item)
                self.tree.Delete(item)
        elif evtId == wx.ID_CLEAR:
            self.tree.DeleteAllItems()


class DirPanel(wx.Panel):

    ID_GOTO_PARENT = wx.NewId()
    ID_GOTO_HOME = wx.NewId()

    def __init__(self, parent):
        wx.Panel.__init__(self, parent)

        agwStyle = aui.AUI_TB_OVERFLOW | aui.AUI_TB_PLAIN_BACKGROUND
        self.tb = aui.AuiToolBar(self, agwStyle=agwStyle)
        self.tb.AddSimpleTool(self.ID_GOTO_PARENT, 'Parent',
                              wx.Bitmap(to_byte(goup_xpm)), 'Parent folder')
        self.tb.AddSimpleTool(self.ID_GOTO_HOME, 'Home',
                              wx.Bitmap(to_byte(home_xpm)), 'Current folder')
        self.tb.Realize()
        self.dirtree = DirTreeCtrl(self,
                                   style=wx.TR_DEFAULT_STYLE
                                   | wx.TR_HAS_VARIABLE_ROW_HEIGHT
                                   | wx.TR_HIDE_ROOT)
        self.dirtree.SetRootDir(os.getcwd())
        self.box = wx.BoxSizer(wx.VERTICAL)
        self.box.Add(self.tb, 0, wx.EXPAND, 0)
        #self.box.Add(wx.StaticLine(self), 0, wx.EXPAND)
        self.box.Add(self.dirtree, 1, wx.EXPAND)

        self.box.Fit(self)
        self.SetSizer(self.box)

        self.Bind(wx.EVT_TOOL, self.OnGotoHome, id=self.ID_GOTO_HOME)
        self.Bind(wx.EVT_TOOL, self.OnGotoParent, id=self.ID_GOTO_PARENT)

        self.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.OnItemActivated,
                  self.dirtree)

    def OnItemActivated(self, event):
        currentItem = event.GetItem()
        filename = self.dirtree.GetItemText(currentItem)
        parentItem = self.dirtree.GetItemParent(currentItem)
        d = self.dirtree.GetItemData(parentItem)
        if isinstance(d, Directory):
            filepath = os.path.join(d.directory, filename)
        else:
            return
        if self.dirtree.ItemHasChildren(currentItem):
            self.dirtree.SetRootDir(filepath)
            return
        (_, ext) = os.path.splitext(filename)
        if ext == '.py':
            dp.send(signal='frame.file_drop', filename=filepath)
        else:
            os.system("start " + filepath)

    def OnGotoHome(self, event):
        root = self.dirtree.GetRootItem()
        if not root:
            return
        d = self.dirtree.GetItemData(root)
        if isinstance(d, Directory):
            if d.directory == os.getcwd():
                return
        self.dirtree.SetRootDir(os.getcwd())

    def OnGotoParent(self, event):
        root = self.dirtree.GetRootItem()
        if not root:
            return
        d = self.dirtree.GetItemData(root)
        if isinstance(d, Directory):
            path = os.path.abspath(os.path.join(d.directory, os.path.pardir))
            if path == d.directory:
                return
            self.dirtree.SetRootDir(path)


class MiscTools(object):
    frame = None

    @classmethod
    def Initialize(cls, frame, **kwargs):
        if cls.frame:
            return
        cls.frame = frame
        if not frame:
            return

        active = kwargs.get('active', True)
        direction = kwargs.get('direction', 'top')
        # history panel
        cls.panelHistory = HistoryPanel(frame)
        dp.send(signal='frame.add_panel',
                panel=cls.panelHistory,
                title="History",
                showhidemenu='View:Panels:Command History',
                active=active,
                direction=direction)
        # help panel
        cls.panelHelp = HelpPanel(frame)
        dp.send(signal='frame.add_panel',
                panel=cls.panelHelp,
                title="Help",
                target='History',
                showhidemenu='View:Panels:Command Help',
                active=active,
                direction=direction)
        # directory panel
        cls.panelDir = DirPanel(frame)
        dp.send(signal='frame.add_panel',
                panel=cls.panelDir,
                title="Browsing",
                target="History",
                showhidemenu='View:Panels:Browsing',
                active=active,
                direction=direction)

        dp.connect(receiver=cls.Uninitialize, signal='frame.exit')

    @classmethod
    def Uninitialize(cls):
        """destroy the module"""
        pass


def bsm_initialize(frame, **kwargs):
    """module initialization"""
    MiscTools.Initialize(frame, **kwargs)



import sys
import traceback
import copy
import six
import wx
import wx.adv
from .validators import *
from .formatters import *
from .enumtype import EnumType

wxEVT_PROP_SELECTED = wx.NewEventType()
wxEVT_PROP_CHANGING = wx.NewEventType()
wxEVT_PROP_CHANGED = wx.NewEventType()
wxEVT_PROP_HIGHLIGHTED = wx.NewEventType()
wxEVT_PROP_RIGHT_CLICK = wx.NewEventType()
wxEVT_PROP_COLLAPSED = wx.NewEventType()
wxEVT_PROP_EXPANDED = wx.NewEventType()
wxEVT_PROP_DOUBLE_CLICK = wx.NewEventType()
wxEVT_PROP_INDENT = wx.NewEventType()
wxEVT_PROP_KEYDOWN = wx.NewEventType()
wxEVT_PROP_RESIZE = wx.NewEventType()
wxEVT_PROP_REFRESH = wx.NewEventType()
wxEVT_PROP_DROP = wx.NewEventType()
wxEVT_PROP_BEGIN_DRAG = wx.NewEventType()

EVT_PROP_SELECTED = wx.PyEventBinder(wxEVT_PROP_SELECTED, 1)
EVT_PROP_CHANGING = wx.PyEventBinder(wxEVT_PROP_CHANGING, 1)
EVT_PROP_CHANGED = wx.PyEventBinder(wxEVT_PROP_CHANGED, 1)
EVT_PROP_HIGHLIGHTED = wx.PyEventBinder(wxEVT_PROP_HIGHLIGHTED, 1)
EVT_PROP_RIGHT_CLICK = wx.PyEventBinder(wxEVT_PROP_RIGHT_CLICK, 1)
EVT_PROP_COLLAPSED = wx.PyEventBinder(wxEVT_PROP_COLLAPSED, 1)
EVT_PROP_EXPANDED = wx.PyEventBinder(wxEVT_PROP_EXPANDED, 1)
EVT_PROP_DOUBLE_CLICK = wx.PyEventBinder(wxEVT_PROP_DOUBLE_CLICK, 1)
EVT_PROP_INDENT = wx.PyEventBinder(wxEVT_PROP_INDENT, 1)
EVT_PROP_KEYDOWN = wx.PyEventBinder(wxEVT_PROP_KEYDOWN, 1)
EVT_PROP_RESIZE = wx.PyEventBinder(wxEVT_PROP_RESIZE, 1)
EVT_PROP_REFRESH = wx.PyEventBinder(wxEVT_PROP_REFRESH, 1)
EVT_PROP_DROP = wx.PyEventBinder(wxEVT_PROP_DROP, 1)
EVT_PROP_BEGIN_DRAG = wx.PyEventBinder(wxEVT_PROP_BEGIN_DRAG, 1)


class Property(object):
    controls = EnumType('default', 'none', 'editbox', 'choice', 'file',
                        'folder', 'slider', 'spin', 'checkbox', 'radiobox',
                        'color', 'date', 'time', 'font')

    def __init__(self, grid, name, label, value):
        self.grid = grid
        self.name = name
        self.label = label
        self.label_tip = ''
        self.value = value
        self.value_tip = ''
        # indicate if data is garbage
        self.value_valid = True
        # -1 to use the default one defined in parent's art provider
        self.title_width = -1
        self.indent = 0
        self.activated = False
        self.enable = True
        self.font_label = None
        self.font_value = None
        self.has_children = False
        self.expanded = True
        self.visible = True
        self.readonly = False
        self.ctrl_type = self.controls.default
        self.window = None
        self.parent = -1
        self.SetTextColor(silent=True)
        self.SetBgColor(silent=True)
        self.min_size = wx.Size(200, 25)
        self.rect = wx.Rect(0, 0, 0, 0)
        # non-overlapping regions
        self.regions = {
            'value': wx.Rect(),
            'label': wx.Rect(),
            'splitter': wx.Rect(),
            'expander': wx.Rect()
        }
        self.show_label_tips = False
        self.show_value_tips = False
        self.separator = False
        self.data = None
        self.formatter = None

    def duplicate(self):
        """
        copy the object

        copy.deepcopy does not work since the object contains pointer to wx
        objects
        """
        p = Property(self.grid, self.name, self.label, self.value)
        p.label_tip = self.label_tip
        p.value_tip = self.value_tip
        p.title_width = self.title_width
        p.indent = self.indent
        p.activated = self.activated
        p.enable = self.enable
        if p.font_label:
            self.font_label = wx.Font(p.font_label)
        if p.font_value:
            self.font_value = wx.Font(p.font_value)
        p.has_children = self.has_children
        p.expanded = self.expanded
        p.visible = self.visible
        p.readonly = self.readonly
        p.ctrl_type = self.ctrl_type
        p.parent = self.parent
        p.SetTextColor(self.text_clr, self.text_clr_sel,
                       self.text_clr_disabled, True)
        p.SetBgColor(self.bg_clr, self.bg_clr_sel, self.bg_clr_disabled, True)
        p.show_label_tips = self.show_label_tips
        p.show_value_tips = self.show_value_tips
        p.separator = self.separator
        p.data = self.data
        p.formatter = copy.copy(p.formatter)
        return p

    def SetGrid(self, grid):
        """set the grid window"""
        self.grid = grid

    def GetGrid(self):
        """return the grid window"""
        return self.grid

    def SetData(self, data):
        self.data = data

    def GetData(self):
        return self.data

    def SetSeparator(self, sep=True, silent=False):
        """set the property to be a separator"""
        if self.separator == sep:
            return
        self.separator = sep
        if not silent:
            self.Refresh()

    def IsSeparator(self):
        """return true if the property is a separator"""
        return self.separator

    def SetControlStyle(self, style):
        """set the control type
        style: default | none | editbox | choice | file | folder | slider |
               spin | checkbox | radiobox | color | date | time
        """
        self.UpdatePropValue()
        self.DestroyControl()
        if style not in self.controls:
            print('Unsupported control style!')
            return False
        self.ctrl_type = self.controls[style]
        return True

    def GetControlStyle(self):
        """return the control type"""
        return self.ctrl_type

    def Enable(self, enable=True, silent=False):
        """enable/disable the property"""
        if self.enable == enable:
            return
        self.enable = enable
        if not silent:
            self.Refresh()

    def IsEnabled(self):
        """return true if the property is enabled"""
        return self.enable

    def SetName(self, name, silent=False):
        """set the name"""
        if self.name == name:
            return
        self.name = name
        if not silent:
            self.Refresh()

    def GetName(self):
        """get the name"""
        return self.name

    def SetLabel(self, label, silent=False):
        """set the label"""
        if self.label == label:
            return
        self.label = label
        if not silent:
            self.Refresh()

    def GetLabel(self):
        """get the label"""
        return self.label

    def SetLabelTip(self, tip):
        """set the label tip"""
        self.label_tip = tip

    def GetLabelTip(self):
        """get the label tip"""
        if self.label_tip:
            return self.label_tip
        return self.GetName()

    def SetLabelFont(self, font, silent=False):
        """set label font"""
        self.font_label = font
        if not silent:
            self.Refresh()

    def GetLabelFont(self):
        """get label font"""
        return self.font_label

    def SetVisible(self, visible=True, silent=False):
        """
        show/hide the property

        The property may be hidden if its parent is in collapsed mode.
        """
        if self.visible == visible:
            return
        self.visible = visible
        if not silent:
            self.Refresh(True)

    def IsVisible(self):
        """return true if the property is visible"""
        return self.visible

    def SetParent(self, prop):
        """set the parent property"""
        if prop and prop.GetIndent() >= self.GetIndent():
            return
        self.parent = prop

    def GetParent(self):
        """return the parent property"""
        return self.parent

    def SetValue(self, value, silent=False):
        """set the value"""
        if self.IsReadonly():
            return False
        if self.value != value:
            self.DestroyControl()
            fmt = self.formatter
            try:
                if fmt and not fmt.validate(fmt.format(value)):
                    return False
            except:
                traceback.print_exc(file=sys.stdout)
                return False
            self.value = value
            if not silent:
                self.Refresh()
            return True
        return False

    def GetValue(self):
        """get the value"""
        return self.value

    def GetValueAsString(self):
        """get the value as string"""
        try:
            if self.formatter and self.GetValueValid():
                # ignore format if data is not valid yet
                return self.formatter.format(self.value)
        except:
            traceback.print_exc(file=sys.stdout)

        return str(self.value)

    def SetValueValid(self, valid):
        self.value_valid = valid

    def GetValueValid(self):
        return self.value_valid

    def SetValueTip(self, tip):
        """set the value tip"""
        self.value_tip = tip

    def GetValueTip(self):
        """get the valuetip"""
        if self.value_tip:
            return self.value_tip
        return self.GetValueAsString()

    def SetValueFont(self, font, silent=False):
        """set value font"""
        self.font_value = font
        if not silent:
            self.Refresh()

    def GetValueFont(self):
        """get value font"""
        return self.font_value

    def SetFormatter(self, formatter):
        if not formatter or not hasattr(formatter, 'validate') or\
           not hasattr(formatter, 'format') or not hasattr(formatter, 'coerce'):
            formatter = None
        self.formatter = formatter

    def GetFormatter(self):
        return self.formatter

    def SetIndent(self, indent, silent=False):
        """set the indent to a positive integer"""
        if indent < 0:
            indent = 0
        if indent == self.indent:
            return
        self.indent = indent
        if not silent:
            self.SendPropEvent(wxEVT_PROP_INDENT)

    def GetIndent(self):
        """get the indent"""
        return self.indent

    def SetExpand(self, expand=True, silent=False):
        """expand/collapse the children"""
        if not self.HasChildren():
            return
        if expand == self.expanded:
            return
        self.expanded = expand
        if silent:
            return
        if self.expanded:
            evt = wxEVT_PROP_EXPANDED
        else:
            evt = wxEVT_PROP_COLLAPSED
        if not silent:
            self.SendPropEvent(evt)

    def IsExpanded(self):
        """return true if the expand/collapse button is expanded"""
        return self.expanded

    def SetHasChildren(self, haschildren, silent=False):
        """Indicate that the property has children"""
        if haschildren == self.has_children:
            return
        self.has_children = haschildren
        if silent:
            return
        if self.expanded:
            evt = wxEVT_PROP_EXPANDED
        else:
            evt = wxEVT_PROP_COLLAPSED
        self.SendPropEvent(evt)

    def HasChildren(self):
        """return true if the property has children"""
        return self.has_children

    def SetActivated(self, activated=True):
        """activate the property"""
        if activated == self.activated:
            return
        self.activated = activated
        if not activated:
            # destroy the control if the property is inactivated
            self.OnTextEnter()
        else:
            self.SendPropEvent(wxEVT_PROP_SELECTED)

    def IsActivated(self):
        """return true if the property is activated"""
        return self.activated

    def SetReadonly(self, readonly=True, silent=False):
        """set the property to readonly"""
        if readonly != self.IsReadonly():
            self.readonly = readonly
            if not silent:
                self.Refresh()

    def IsReadonly(self):
        """return true if the property is readonly"""
        return self.readonly

    def SetTextColor(self,
                     clr=None,
                     clr_sel=None,
                     clr_disabled=None,
                     silent=False):
        """
        set the text colors

        All values are string. If the value is None, the color will reset to
        default.
        """
        self.text_clr = clr
        self.text_clr_sel = clr_sel
        self.text_clr_disabled = clr_disabled

        if not silent:
            self.Refresh()

    def GetTextColor(self):
        """get the text colors"""
        return (self.text_clr, self.text_clr_sel, self.text_clr_disabled)

    def SetBgColor(self,
                   clr=None,
                   clr_sel=None,
                   clr_disabled=None,
                   silent=False):
        """
        set the background colors

        All values are string. If the value is None, the color will reset to
        default.
        """
        self.bg_clr = clr
        self.bg_clr_sel = clr_sel
        self.bg_clr_disabled = clr_disabled

        if not silent:
            self.Refresh()

    def GetBgColor(self):
        """get the background colors"""
        return (self.bg_clr, self.bg_clr_sel, self.bg_clr_disabled)

    def SetTitleWidth(self, width):
        """set the title width"""
        self.title_width = width

    def GetTitleWidth(self):
        """return the width"""
        return self.title_width

    def SetRect(self, rc):
        """set the prop rect"""
        if self.rect != rc:
            self.rect = wx.Rect(*rc)
            # redraw the item
            self.Refresh()
            wx.CallAfter(self.LayoutControl)

    def GetRect(self):
        """return the prop rect"""
        return wx.Rect(*self.rect)

    def SetMinSize(self, size, silent=False):
        """set the min size"""
        if self.min_size != size:
            self.min_size = wx.Size(*size)
            if not silent:
                self.Resize()

    def GetMinSize(self):
        """return the min size"""
        if self.window:
            size = self.window.GetSize()
            sz = self.min_size
            size.y = max(sz.y, size.y)
            return size
        return wx.Size(*self.min_size)

    def GetSize(self):
        """return the current size"""
        return self.rect.GetSize()

    def GetShowLabelTips(self):
        """return whether label tooltip is allowed"""
        return self.show_label_tips

    def GetShowValueTips(self):
        """return whether value tooltip is allowed"""
        return self.show_value_tips

    def HitTest(self, pt):
        """find the mouse position relative to the property"""
        # bottom edge
        rc = wx.Rect(*self.rect)
        rc.SetTop(rc.bottom - 2)
        if rc.Contains(pt):
            return 'bottom_edge'
        # top edge
        rc = wx.Rect(*self.rect)
        rc.SetBottom(rc.top + 2)
        if rc.Contains(pt):
            return 'top_edge'

        for k, v in six.iteritems(self.regions):
            if v.Contains(pt):
                return k

        return None

    def OnMouseDown(self, pt):
        ht = self.HitTest(pt)
        # click on the expand buttons? expand it?
        if self.HasChildren() and ht == 'expander':
            self.SetExpand(not self.expanded)

        if ht == 'value':
            if not self.IsReadonly() and self.IsActivated():
                self.CreateControl()
        else:
            self.UpdatePropValue()
            self.DestroyControl()

        return ht

    def OnMouseUp(self, pt):
        ht = self.HitTest(pt)
        return ht

    def OnMouseDoubleClick(self, pt):
        ht = self.HitTest(pt)

        if self.IsEnabled():
            if ht == 'value':
                if not self.IsReadonly():
                    self.CreateControl()
            else:
                self.UpdatePropValue()
                self.DestroyControl()

            if ht == 'expander':
                self.SetExpand(not self.expanded)

            self.SendPropEvent(wxEVT_PROP_DOUBLE_CLICK)

        return ht

    def OnMouseRightClick(self, pt):
        ht = self.HitTest(pt)

        if self.IsEnabled():
            # destroy the control when the mouse moves out
            if ht is None:
                self.UpdatePropValue()
                self.DestroyControl()
            self.SendPropEvent(wxEVT_PROP_RIGHT_CLICK)

        return ht

    def OnMouseMove(self, pt):
        ht = self.HitTest(pt)
        return ht

    def OnTextEnter(self):
        self.UpdatePropValue()
        self.DestroyControl()
        self.Refresh()

    def CreateControl(self):
        """create the control"""
        if self.window != None or self.IsSeparator():
            return
        style = self.ctrl_type
        if style == self.controls.default:
            style = self.controls.editbox
            if isinstance(self.formatter, EnumFormatter) or \
               isinstance(self.formatter, ChoiceFormatter):
                style = self.controls.choice
            elif isinstance(self.formatter, BoolFormatter):
                style = self.controls.checkbox
            elif isinstance(self.formatter, PathFormatter):
                if self.formatter.types == 'file':
                    style = self.controls.file
                elif self.formatter.types == 'folder':
                    style = self.controls.folder
            elif isinstance(self.formatter, ColorFormatter):
                style = self.controls.color
            elif isinstance(self.formatter, FontFormatter):
                style = self.controls.font
            elif isinstance(self.formatter, DateFormatter):
                style = self.controls.date
                if isinstance(self.formatter, TimeFormatter):
                    style = self.controls.time
        win = None
        if style == self.controls.editbox:
            style = wx.TE_PROCESS_ENTER
            sz = self.GetMinSize()
            if sz.y > 50:
                style = wx.TE_MULTILINE
            win = wx.TextCtrl(self.grid,
                              wx.ID_ANY,
                              self.GetValueAsString(),
                              style=style)
            if self.formatter:
                validator = TextValidator(self, 'value', self.formatter, False,
                                          None)
                win.SetValidator(validator)

            if style & wx.TE_PROCESS_ENTER:
                win.Bind(wx.EVT_TEXT_ENTER, self.OnPropTextEnter)

        elif style == self.controls.choice:
            win = wx.Choice(self.grid, wx.ID_ANY)
            if self.formatter:
                validator = SelectorValidator(self, 'value', self.formatter,
                                              True)
                win.SetValidator(validator)

        elif style == self.controls.file:
            win = wx.Button(self.grid, wx.ID_ANY, self.GetValueAsString())
            win.Bind(wx.EVT_BUTTON, self.OnSelectFile)

        elif style == self.controls.folder:
            win = wx.Button(self.grid, wx.ID_ANY, self.GetValueAsString())
            win.Bind(wx.EVT_BUTTON, self.OnSelectFolder)

        elif style == self.controls.slider:
            win = wx.Slider(self.grid,
                            wx.ID_ANY,
                            value=int(self.value),
                            style=wx.SL_LABELS | wx.SL_HORIZONTAL | wx.SL_TOP)
            if self.formatter:
                validator = SpinSliderValidator(self, 'value', self.formatter,
                                                True)
                win.SetValidator(validator)

        elif style == self.controls.spin:
            win = wx.SpinCtrl(self.grid,
                              wx.ID_ANY,
                              value=str(self.value),
                              style=wx.SP_ARROW_KEYS)
            if self.formatter:
                validator = SpinSliderValidator(self, 'value', self.formatter,
                                                True)
                win.SetValidator(validator)

        elif style == self.controls.checkbox:
            win = wx.CheckBox(self.grid, wx.ID_ANY)
            win.SetValue(int(self.value) != 0)

        elif style == self.controls.radiobox:
            choices = []
            if self.formatter and hasattr(self.formatter, 'validValues'):
                choices = [x[1] for x in self.formatter.validValues()]

            win = wx.RadioBox(self.grid, wx.ID_ANY, "", wx.DefaultPosition,
                              wx.DefaultSize, choices, 5, wx.RA_SPECIFY_COLS)

            if self.formatter:
                validator = RadioBoxValidator(self, 'value', self.formatter,
                                              True)
                win.SetValidator(validator)

        elif style == self.controls.color:
            win = wx.ColourPickerCtrl(self.grid,
                                      wx.ID_ANY,
                                      wx.BLACK,
                                      style=wx.CLRP_DEFAULT_STYLE
                                      | wx.CLRP_SHOW_LABEL)
            try:
                win.SetColour(self.value)
            except ValueError:
                pass
        elif style == self.controls.font:
            win = wx.FontPickerCtrl(self.grid, wx.ID_ANY)
            try:
                if self.value:
                    win.SetSelectedFont(self.value)
            except ValueError:
                pass

        elif style == self.controls.date:
            win = wx.adv.DatePickerCtrl(self.grid, wx.ID_ANY)
            try:
                win.SetValue(self.value)
            except ValueError:
                pass

        elif style == self.controls.time:
            win = wx.adv.TimePickerCtrl(self.grid, wx.ID_ANY)
            try:
                win.SetValue(self.value)
            except ValueError:
                pass
        if win:
            if win.GetValidator():
                win.GetValidator().TransferToWindow()
            self.window = win
            # the window size may be larger than the value rect, notify parent
            # grid to update it
            self.Resize()
            self.LayoutControl()
            self.window.SetFocus()
            self.window.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)

    def OnPropTextEnter(self, evt):
        """send when the enter key is pressed in the property control window"""
        if self.window:
            self.OnTextEnter()

    def OnKillFocus(self, evt):
        # destroy the control if it loses focus. Wait until the event has been
        # processed; otherwise, it may crash.
        evt.Skip()
        wnd = evt.GetWindow()
        while wnd:
            if wnd.GetParent() == self.window:
                return
            wnd = wnd.GetParent()
        # color window does not work on Mac
        #wx.CallAfter(self.OnTextEnter)

    def OnSelectFile(self, evt):
        style = wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
        dlg = wx.FileDialog(self.grid, "Choose a file",
                            self.GetValueAsString(), "", "*.*", style)
        if dlg.ShowModal() == wx.ID_OK:
            self.SetValue(dlg.GetPath())
        dlg.Destroy()

    def OnSelectFolder(self, evt):
        dlg = wx.DirDialog(self.grid, "Choose input directory",
                           self.GetValueAsString(),
                           wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)
        if dlg.ShowModal() == wx.ID_OK:
            self.SetValue(dlg.GetPath())
        dlg.Destroy()

    def LayoutControl(self):
        """re-positioning the control"""
        if self.window is None:
            return
        rc = self.grid.GetScrolledRect(wx.Rect(*self.regions['value']))
        self.window.SetSize(rc.GetSize())
        self.window.Move(rc.GetTopLeft())

    def DestroyControl(self):
        """destroy the value setting control"""
        if self.window:
            self.window.Show(False)
            self.window.Destroy()
            self.window = None
            self.Resize()
            return True
        return False

    def UpdatePropValue(self):
        """update the value"""
        if self.window is None:
            return False
        validator = self.window.GetValidator()
        if validator:
            validator.TransferFromWindow()
            return

        value_old = self.value
        value = None
        if isinstance(self.window, wx.TextCtrl):
            value = self.window.GetValue()

        elif isinstance(self.window, wx.Button):
            value = self.window.GetLabel()

        elif isinstance(self.window, wx.RadioBox) or\
             isinstance(self.window, wx.Choice):
            sel = self.window.GetSelection()
            if sel >= 0 and sel < self.window.GetCount():
                value = self.window.GetString(sel)

        elif isinstance(self.window, wx.Slider):
            value = self.window.GetValue()

        elif isinstance(self.window, wx.SpinCtrl):
            value = self.window.GetValue()

        elif isinstance(self.window, wx.CheckBox):
            value = self.window.GetValue()

        elif isinstance(self.window, wx.ColourPickerCtrl):
            clr = self.window.GetColour()
            value = clr.GetAsString(wx.C2S_HTML_SYNTAX)

        elif isinstance(self.window, wx.FontPickerCtrl):
            font = self.window.GetSelectedFont()
            value = font.GetNativeFontInfoDesc()

        elif isinstance(self.window, wx.adv.DatePickerCtrl) or\
             isinstance(self.window, wx.adv.TimePickerCtrl):
            value = self.window.GetValue()
            if self.formatter:
                value = self.formatter.format(value)
        try:
            if self.formatter:
                if self.formatter.validate(str(value)):
                    value = self.formatter.coerce(str(value))
                else:
                    return False
            elif self.value is not None:
                value = type(self.value)(value)
            self.SetValue(value, silent=True)
        except ValueError:
            traceback.print_exc(file=sys.stdout)
            return False

        if self.SendPropEvent(wxEVT_PROP_CHANGING):
            self.SendPropEvent(wxEVT_PROP_CHANGED)
            self.Refresh()
            return True
        else:
            #the parent rejects the operation, restore the original value
            self.SetValue(value_old)
            return False

    def SendPropEvent(self, event):
        """ send property grid event to parent"""
        win = self.GetGrid()
        evt = PropertyEvent(event)
        evt.SetProperty(self)
        evt.SetEventObject(win)
        evt_handler = win.GetEventHandler()

        if evt_handler.ProcessEvent(evt):
            return not evt.GetVeto()
        return False

    def Resize(self):
        """notify parent grid the size needs to be updated"""
        self.SendPropEvent(wxEVT_PROP_RESIZE)

    def Refresh(self, force=False):
        """notify the parent to redraw the property"""
        if self.IsVisible() or force:
            self.SendPropEvent(wxEVT_PROP_REFRESH)


class PropertyEvent(wx.PyCommandEvent):
    def __init__(self, commandType, id=0):
        wx.PyCommandEvent.__init__(self, commandType, id)
        self.prop = None
        self.veto = False

    def GetProperty(self):
        """return the attached Property"""
        return self.prop

    def SetProperty(self, prop):
        """attach the Property instance"""
        if isinstance(prop, Property):
            self.prop = prop

    def Veto(self, veto=True):
        """refuse the event"""
        self.veto = veto

    def GetVeto(self):
        """return whether the event is refused"""
        return self.veto



import inspect
import six
import wx
import wx.py.dispatcher as dp
import wx.lib.mixins.listctrl as listmix
import wx.lib.agw.aui as aui
from .bsmxpm import (run_svg, run_grey_svg, step_over_svg, step_over_grey_svg, step_into_svg, \
                     step_into_grey_svg, step_out_svg, step_out_grey_svg, step_over_svg, stop_svg, stop_grey_svg)

from .utility import svg_to_bitmap

class StackListCtrl(wx.ListCtrl, listmix.ListCtrlAutoWidthMixin,
                    listmix.ListRowHighlighter):
    def __init__(self, *args, **kwargs):
        wx.ListCtrl.__init__(self, *args, **kwargs)

        listmix.ListCtrlAutoWidthMixin.__init__(self)
        listmix.ListRowHighlighter.__init__(self, mode=listmix.HIGHLIGHT_ODD)
        self.SetHighlightColor(wx.Colour(240, 240, 250))


class StackPanel(wx.Panel):
    def __init__(self, parent):
        wx.Panel.__init__(self, parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.listctrl = StackListCtrl(self,
                                      style=wx.LC_REPORT
                                      | wx.BORDER_NONE
                                      | wx.LC_EDIT_LABELS | wx.LC_VRULES
                                      | wx.LC_HRULES | wx.LC_SINGLE_SEL)
        # | wx.BORDER_SUNKEN
        # | wx.LC_SORT_ASCENDING
        # | wx.LC_NO_HEADER
        self.listctrl.InsertColumn(0, 'Name')
        self.listctrl.InsertColumn(1, 'Line')
        self.listctrl.InsertColumn(2, 'File')
        sizer.Add(self.listctrl, 1, wx.ALL | wx.EXPAND, 5)
        self.SetSizer(sizer)
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.OnItemActivated,
                  self.listctrl)
        dp.connect(self.OnDebugEnded, 'debugger.ended')
        dp.connect(self.OnDebugUpdateScopes, 'debugger.update_scopes')
        dp.connect(self.OnDebugUpdateScopes, 'debugger.paused')

    def Destroy(self):
        dp.disconnect(self.OnDebugEnded, 'debugger.ended')
        dp.disconnect(self.OnDebugUpdateScopes, 'debugger.update_scopes')
        dp.disconnect(self.OnDebugUpdateScopes, 'debugger.paused')
        super(StackPanel, self).Destroy()

    def OnDebugEnded(self):
        """debugger is ended"""
        # clear the scopes
        self.listctrl.DeleteAllItems()

    def OnDebugUpdateScopes(self):
        """debugger changes the scope"""
        self.listctrl.DeleteAllItems()
        resp = dp.send(signal='debugger.get_status')
        if not resp or not resp[0][1]:
            return
        status = resp[0][1]
        frames = status['frames']
        level = status['active_scope']
        if frames is not None:
            for frame in frames:
                name = frame.f_code.co_name
                filename = inspect.getsourcefile(frame) or inspect.getfile(
                    frame)
                lineno = frame.f_lineno
                index = self.listctrl.InsertItem(six.MAXSIZE, name)
                self.listctrl.SetItem(index, 2, filename)
                self.listctrl.SetItem(index, 1, '%d' % lineno)
        if level >= 0 and level < self.listctrl.GetItemCount():
            self.listctrl.SetItemTextColour(level, 'blue')
        self.listctrl.RefreshRows()

    def OnItemActivated(self, event):
        currentItem = event.m_itemIndex
        filename = self.listctrl.GetItem(currentItem, 2).GetText()
        lineno = self.listctrl.GetItem(currentItem, 1).GetText()
        # open the script first
        dp.send(signal='frame.file_drop',
                filename=filename,
                lineno=int(lineno))
        # ask the debugger to trigger the update scope event to set mark
        dp.send(signal='debugger.set_scope', level=currentItem)


class DebugTool(object):
    isInitialized = False
    frame = None
    showStackPanel = True

    @classmethod
    def Initialize(cls, frame, **kwargs):
        if cls.isInitialized:
            return
        cls.isInitialized = True
        cls.frame = frame
        # stack panel
        cls.panelStack = StackPanel(frame)
        dp.send('frame.add_panel',
                panel=cls.panelStack,
                title="Call Stack",
                active=False,
                showhidemenu='View:Panels:Call Stack')

        # debugger toolbar
        dp.send('frame.add_menu',
                path='Tools:Debug',
                rxsignal='',
                kind='Popup')
        cls.tbDebug = aui.AuiToolBar(frame, style=wx.TB_FLAT | wx.TB_HORIZONTAL)
        items = (
            ('Run\tF5', 'resume', run_svg, run_grey_svg, 'paused'),
            ('Stop\tShift-F5', 'stop', stop_svg, stop_grey_svg, 'paused'),
            ('Step\tF10', 'step', step_over_svg, step_over_grey_svg, 'paused'),
            ('Step Into\tF11', 'step_into', step_into_svg, step_into_grey_svg, 'can_stepin'),
            ('Step Out\tShift-F11', 'step_out', step_out_svg, step_out_grey_svg, 'can_stepout'),
        )
        cls.menus = {}
        for label, signal, xpm, xpm_grey, status in items:
            resp = dp.send('frame.add_menu',
                           path='Tools:Debug:' + label,
                           rxsignal='debugger.' + signal,
                           updatesignal='debugtool.updateui')
            if not resp:
                continue
            cls.menus[resp[0][1]] = status
            cls.tbDebug.AddTool(resp[0][1], label, svg_to_bitmap(xpm),
                                svg_to_bitmap(xpm_grey), wx.ITEM_NORMAL, label)
        cls.tbDebug.Realize()

        dp.send('frame.add_panel',
                panel=cls.tbDebug,
                title='Debugger',
                active=False,
                paneInfo=aui.AuiPaneInfo().Name('debugger').Caption(
                    'Debugger').ToolbarPane().Top(),
                showhidemenu='View:Toolbars:Debugger')
        dp.connect(cls.OnUpdateMenuUI, 'debugtool.updateui')
        dp.connect(cls.OnDebugPaused, 'debugger.paused')
        dp.connect(cls.OnDebugEnded, 'debugger.ended')
        dp.connect(cls.Uninitialize, 'frame.exit')

    @classmethod
    def Uninitialize(cls):
        """destroy the module"""
        pass

    @classmethod
    def OnDebugPaused(cls):
        """update the debug toolbar status"""
        resp = dp.send('debugger.get_status')
        if not resp or not resp[0][1]:
            return
        status = resp[0][1]
        paused = status['paused']
        for k, s in six.iteritems(cls.menus):
            cls.tbDebug.EnableTool(k, paused and status.get(s, False))
        cls.tbDebug.Refresh(False)
        if paused and not cls.tbDebug.IsShown():
            dp.send('frame.show_panel', panel=cls.tbDebug)

        if cls.showStackPanel and paused and not cls.panelStack.IsShown():
            dp.send('frame.show_panel', panel=cls.panelStack)
            # allow the use to hide the Stack panel
            cls.showStackPanel = False

    @classmethod
    def OnDebugEnded(cls):
        """debugger is ended"""
        # disable and hide the debugger toolbar
        for k in six.iterkeys(cls.menus):
            cls.tbDebug.EnableTool(k, False)
        cls.tbDebug.Refresh(False)

        # hide the debugger toolbar and Stack panel
        dp.send('frame.show_panel', panel=cls.tbDebug, show=False)
        dp.send('frame.show_panel', panel=cls.panelStack, show=False)
        # show the Stack panel next time
        cls.showStackPanel = True

    @classmethod
    def OnUpdateMenuUI(cls, event):
        """update the debugger toolbar"""
        eid = event.GetId()
        resp = dp.send('debugger.get_status')
        enable = False
        if resp and resp[0][1]:
            status = resp[0][1]
            paused = status['paused']

            s = cls.menus.get(eid, 'paused')

            enable = paused and status[s]
        event.Enable(enable)


def bsm_initialize(frame, **kwargs):
    """module initialization"""
    DebugTool.Initialize(frame, **kwargs)



# -*- coding: utf-8 -*-
__license__ = """Copyright (c) 2008-2010, Toni Ruža, All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 'AS IS'
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE."""

__author__ = u"Toni Ruža <gmr.gaf@gmail.com>"
__url__ = "http://bitbucket.org/raz/wxautocompletectrl"

import six
import wx
import wx.html


class SuggestionsPopup(wx.Frame):
    def __init__(self, parent):
        wx.Frame.__init__(self,
                          parent,
                          style=wx.NO_BORDER | wx.FRAME_TOOL_WINDOW
                          | wx.FRAME_NO_TASKBAR | wx.FRAME_FLOAT_ON_PARENT
                          | wx.STAY_ON_TOP)
        self._suggestions = self._listbox(self)
        self._suggestions.SetItemCount(0)
        self._unformated_suggestions = None

    class _listbox(wx.html.HtmlListBox):
        items = None

        def OnGetItem(self, n):
            return self.items[n]

    def SetSuggestions(self, suggestions, unformated_suggestions):
        self._suggestions.items = suggestions
        self._suggestions.SetItemCount(len(suggestions))
        self._suggestions.SetSelection(0)
        self._suggestions.Refresh()
        self._unformated_suggestions = unformated_suggestions

    def CursorUp(self):
        selection = self._suggestions.GetSelection()
        if selection > 0:
            self._suggestions.SetSelection(selection - 1)

    def CursorDown(self):
        selection = self._suggestions.GetSelection()
        last = self._suggestions.GetItemCount() - 1
        if selection < last:
            self._suggestions.SetSelection(selection + 1)

    def CursorHome(self):
        if self.IsShown():
            self._suggestions.SetSelection(0)

    def CursorEnd(self):
        if self.IsShown():
            self._suggestions.SetSelection(self._suggestions.GetItemCount() -
                                           1)

    def GetSelectedSuggestion(self):
        return self._unformated_suggestions[self._suggestions.GetSelection()]

    def GetSuggestion(self, n):
        return self._unformated_suggestions[n]


class AutocompleteTextCtrl(wx.TextCtrl):
    def __init__(self,
                 parent,
                 height=300,
                 completer=None,
                 multiline=False,
                 frequency=250,
                 value=""):
        style = wx.TE_PROCESS_ENTER
        if multiline:
            style = style | wx.TE_MULTILINE
        wx.TextCtrl.__init__(self, parent, value=value, style=style)
        self.height = height
        self.frequency = frequency
        if completer:
            self.SetCompleter(completer)
        self.queued_popup = False
        self.skip_event = False
        self._string = value

    def SetCompleter(self, completer):
        """
        Initializes the autocompletion. The 'completer' has to be a function
        with one argument (the current value of the control, ie. the query)
        and it has to return two lists: formated (html) and unformated
        suggestions.
        """
        self.completer = completer

        frame = self.GetParent()
        while frame and (not isinstance(frame, wx.Frame)) and\
              (not isinstance(frame, wx.Dialog)):
            frame = frame.GetParent()

        self.popup = SuggestionsPopup(frame)

        frame.Bind(wx.EVT_MOVE, self.OnMove)
        self.Bind(wx.EVT_TEXT, self.OnTextUpdate)
        self.Bind(wx.EVT_SIZE, self.OnSizeChange)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.popup._suggestions.Bind(wx.EVT_LEFT_DOWN,
                                     self.OnSuggestionClicked)
        self.popup._suggestions.Bind(wx.EVT_KEY_DOWN, self.OnSuggestionKeyDown)

    def AdjustPopupPosition(self):
        self.popup.Position = self.ClientToScreen((0, self.GetSize().GetHeight())).Get()

    def OnMove(self, event):
        self.AdjustPopupPosition()
        event.Skip()

    def OnTextUpdate(self, event):
        # only show the popup when the text has changed; so type in 'Enter'
        # will not bring the popup
        if event.GetString() != self._string:
            self._string = event.GetString()
            if self.skip_event:
                self.skip_event = False
            elif not self.queued_popup:
                wx.CallLater(self.frequency, self.AutoComplete)
                self.queued_popup = True
        event.Skip()

    def AutoComplete(self, *args, **kwargs):
        self.queued_popup = False
        if self.Value != "":
            formated, unformated, offset = self.completer(self.Value)
            if formated:
                self.auto_comp_offset = offset
                self.popup.SetSuggestions(formated, unformated)
                self.AdjustPopupPosition()
                self.Unbind(wx.EVT_KILL_FOCUS)
                self.popup.ShowWithoutActivating()
                self.SetFocus()
                # in linux, SetFocus may select the whole string, de-select it
                self.SelectNone()
                self.Bind(wx.EVT_KILL_FOCUS, self.OnKillFocus)
            else:
                self.popup.Hide()
        else:
            self.popup.Hide()

    def OnSizeChange(self, event):
        self.popup.Size = (self.Size[0], self.height)
        event.Skip()

    def OnKeyDown(self, event):
        key = event.GetKeyCode()

        if key == wx.WXK_UP:
            self.popup.CursorUp()
            return

        elif key == wx.WXK_DOWN:
            self.popup.CursorDown()
            return

        elif key in (wx.WXK_RETURN, wx.WXK_NUMPAD_ENTER) and self.popup.Shown:
            self.skip_event = True

            txt = self.popup.GetSelectedSuggestion()
            start = end = self.GetLastPosition()
            if self.auto_comp_offset:
                start -= self.auto_comp_offset
            self.Replace(start, end, txt)
            self.SetInsertionPointEnd()
            self.popup.Hide()
            return

        elif key == wx.WXK_HOME:
            self.popup.CursorHome()

        elif key == wx.WXK_END:
            self.popup.CursorEnd()

        elif event.ControlDown() and six.unichr(key).lower() == "a":
            self.SelectAll()

        elif key == wx.WXK_ESCAPE:
            self.popup.Hide()
            return

        event.Skip()

    def OnSuggestionClicked(self, event):
        self.skip_event = True
        n = self.popup._suggestions.HitTest(event.Position)
        self.Value = self.popup.GetSuggestion(n)
        self.SetInsertionPointEnd()
        wx.CallAfter(self.SetFocus)
        event.Skip()

    def OnSuggestionKeyDown(self, event):
        key = event.GetKeyCode()
        if key in (wx.WXK_RETURN, wx.WXK_NUMPAD_ENTER):
            self.skip_event = True
            self.SetValue(self.popup.GetSelectedSuggestion())
            self.SetInsertionPointEnd()
            self.popup.Hide()
        event.Skip()

    def OnKillFocus(self, event):
        if not self.popup.IsActive():
            self.popup.Hide()
        event.Skip()



import sys
import os
import ctypes
import six
from ..cparser import cparser, cwrapper

#CParserFunc = cparser.caching.parse
# If we don't want to use caching:
CParserFunc = cparser.parse


def init_dll(dll, header):
    dll = ctypes.cdll.LoadLibrary(dll)
    parsedState = CParserFunc(header)

    wrapper = cwrapper.CWrapper()
    wrapper.register(parsedState, dll)
    return wrapper.wrapped


def callback(proto, fun):
    SIM_CALLBACK = ctypes.CFUNCTYPE(proto.restype, *proto.argtypes)
    return SIM_CALLBACK(fun)


class SStructWrapper(ctypes.Structure):
    def __init__(self, obj, *args, **kwargs):
        ctypes.Structure.__init__(self, *args, **kwargs)
        self._object = obj

    def __call__(self):
        return self._object

    def is_type(self, obj_type):
        return isinstance(self._object, obj_type)

    def set_object(self, item, value):
        if isinstance(item, six.string_types):
            if hasattr(self._object, item):
                v = getattr(self._object, item)
                if isinstance(v, ctypes.Array) and \
                   isinstance(v, ctypes.ARRAY(ctypes.c_byte, len(v))):
                    setattr(self._object, item, (ctypes.c_byte * len(v))(
                        *bytearray(str(value).encode())))
                else:
                    setattr(self._object, item, value)
                return True
        return False

    def __getattr__(self, item):
        if item == '_object':
            raise AttributeError()
        if hasattr(self._object, item):
            v = getattr(self._object, item)
            if isinstance(v, ctypes.Array) and \
               isinstance(v, ctypes.ARRAY(ctypes.c_byte, len(v))):
                return ctypes.cast(v, ctypes.c_char_p).value.decode()
            return v
        raise AttributeError()

    def __setattr__(self, item, val):
        if hasattr(self, '_object'):
            if self.set_object(item, val):
                return
        super(SStructWrapper, self).__setattr__(item, val)

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            if hasattr(self._object, item):
                v = getattr(self._object, item)
                if isinstance(v, ctypes.Array) and \
                   isinstance(v, ctypes.ARRAY(ctypes.c_byte, len(v))):
                    return ctypes.cast(v, ctypes.c_char_p).value
                return v
        return None

    def __setitem__(self, item, val):
        return self.set_object(item, val)



import six
import wx
import wx.lib.agw.aui as aui
import wx.py.dispatcher as dp


class AuiManagerPlus(aui.AuiManager):
    def __init__(self, managed_window=None, agwFlags=None):
        aui.AuiManager.__init__(self,
                                managed_window=managed_window,
                                agwFlags=agwFlags)

    def RefreshPaneCaption(self, window):
        """used to rename a panel"""
        if window is None:
            return
        pane = self.GetPane(window)
        if not pane.IsOk():
            return
        parent = window.GetParent()
        if parent is None:
            return
        if pane.IsNotebookPage() and isinstance(parent,
                                                aui.auibook.AuiNotebook):
            idx = parent.GetPageIndex(window)
            parent.SetPageText(idx, pane.caption)
            if idx == parent.GetSelection():
                page_info = self.GetPane(parent)
                page_info.Caption(pane.caption)
        self.RefreshCaptions()
        parent.Update()

    def CreateNotebook(self):
        """
        Creates an automatic :class:`~wx.lib.agw.aui.auibook.AuiNotebook` when a pane is docked on
        top of another pane.
        """
        nb = super(AuiManagerPlus, self).CreateNotebook()
        # the default size is (0, 0) from previous step, set it to (100, 100) so it will not assert
        # wxpython/ext/wxWidgets/src/gtk/bitmap.cpp(539): assert ""width > 0 &&
        # height > 0"" failed in Create(): invalid bitmap size
        nb.SetSize(100, 100)
        return nb

    def UpdateNotebook(self):
        """ Updates the automatic :class:`~lib.agw.aui.auibook.AuiNotebook` in
        the layout (if any exists). """

        super(AuiManagerPlus, self).UpdateNotebook()

        # set the notebook minsize to be the largest one of its children page.
        # otherwise, when you drag the notebook, its children page's size may
        # be zero in one dimension, which may cause some assert (e.g., invalid
        # bitmap size as show above)

        for nb in self._notebooks:
            pages = nb.GetPageCount()
            nb_pane = self.GetPane(nb)
            min_x, min_y = nb_pane.min_size
            # Check each tab ...
            for page in range(pages):
                window = nb.GetPage(page)
                page_pane = self.GetPane(window)
                min_x = max(page_pane.min_size.x, min_x)
                min_y = max(page_pane.min_size.y, min_y)
            nb_pane.MinSize(min_x, min_y)


class FramePlus(wx.Frame):
    def __init__(self,
                 parent,
                 id=wx.ID_ANY,
                 title=u'bsmedit',
                 pos=wx.DefaultPosition,
                 size=wx.Size(800, 600),
                 style=wx.DEFAULT_FRAME_STYLE | wx.TAB_TRAVERSAL):
        wx.Frame.__init__(self,
                          parent,
                          id=id,
                          title=title,
                          pos=pos,
                          size=size,
                          style=style)
        self._mgr = AuiManagerPlus()
        self._mgr.SetManagedWindow(self)
        self.menuAddon = {}
        self.paneAddon = {}
        self._pane_num = 0
        dp.connect(self.AddMenu, 'frame.add_menu')
        dp.connect(self.DeleteMenu, 'frame.delete_menu')
        dp.connect(self.AddPanel, 'frame.add_panel')
        dp.connect(self.DeletePanel, 'frame.delete_panel')
        dp.connect(self.ShowPanel, 'frame.show_panel')
        dp.connect(self.TogglePanel, 'frame.check_menu')
        dp.connect(self.UpdateMenu, 'frame.update_menu')
        self.Bind(wx.EVT_CLOSE, self.OnClose)

    def OnClose(self, event):
        self._mgr.UnInit()
        del self._mgr
        event.Skip()

    def GetMenu(self, pathlist, autocreate=False):
        """
        find the menu item.

        if autocreate is True, then recursive submenu creation.
        """
        if not pathlist:
            return None
        # the top level menu
        menuidx = self.GetMenuBar().FindMenu(pathlist[0])
        if menuidx == wx.NOT_FOUND:
            if autocreate:
                self.GetMenuBar().Append(wx.Menu(), pathlist[0])
                menuidx = self.GetMenuBar().FindMenu(pathlist[0])
            else:
                return None
        menuitem = self.GetMenuBar().GetMenu(menuidx)
        for p in pathlist[1:]:
            if menuitem is None:
                return None
            for m in six.moves.range(menuitem.GetMenuItemCount()):
                child = menuitem.FindItemByPosition(m)
                if not child.IsSubMenu():
                    continue
                stritem = child.GetItemLabelText()
                stritem = stritem.split('\t')[0]
                if stritem == p.split('\t')[0]:
                    menuitem = child.GetSubMenu()
                    break
            else:
                if autocreate:
                    child = self._append_menu(menuitem, p, kind='Popup')
                    menuitem = child.GetSubMenu()
                else:
                    return None
        return menuitem

    def _append_menu(self,
                     menu,
                     label,
                     id=None,
                     rxsignal=None,
                     updatesignal=None,
                     kind='Normal'):
        """
        append an item to menu.
            kind: 'Separator', 'Normal', 'Check', 'Radio', 'Popup'
        """
        if menu is None:
            return None

        if kind == 'Separator':
            return menu.AppendSeparator()
        elif kind == 'Popup':
            return menu.AppendSubMenu(wx.Menu(), label)
        else:
            newid = id
            if newid is None:
                newid = wx.NewId()
            if kind == 'Normal':
                newitem = wx.MenuItem(menu,
                                      newid,
                                      label,
                                      label,
                                      kind=wx.ITEM_NORMAL)
            elif kind == 'Check':
                newitem = wx.MenuItem(menu,
                                      newid,
                                      label,
                                      label,
                                      kind=wx.ITEM_CHECK)
            elif kind == 'Radio':
                newitem = wx.MenuItem(menu,
                                      newid,
                                      label,
                                      label,
                                      kind=wx.ITEM_RADIO)
            self.menuAddon[newid] = (rxsignal, updatesignal)
            child = menu.Append(newitem)
            self.Bind(wx.EVT_MENU, self.OnMenuAddOn, id=newid)
            if updatesignal:
                self.Bind(wx.EVT_UPDATE_UI, self.OnMenuCmdUI, id=newid)
            return child
        return None

    def AddMenu(self,
                path,
                id=None,
                rxsignal=None,
                updatesignal=None,
                kind='Normal',
                autocreate=False):
        """
        add the item to menubar.
            path: e.g., New:Open:Figure

            kind: 'Separator', 'Normal', 'Check', 'Radio', 'Popup'
        """

        paths = path.split(':')
        menu = None

        if len(paths) == 1:
            # top level menu
            return self.GetMenuBar().Append(wx.Menu(), paths[0])
        elif len(paths) > 1:
            menu = self.GetMenu(paths[:-1], autocreate)
            child = self._append_menu(menu, paths[-1], id, rxsignal,
                                      updatesignal, kind)
            if child:
                return child.GetId()
        return wx.NOT_FOUND

    def DeleteMenu(self, path, id=None):
        """delete the menu item"""
        pathlist = path.split(':')
        menuitem = self.GetMenu(pathlist[:-1])
        if menuitem is None:
            return False
        if id is None:
            # delete a submenu
            for m in six.moves.range(menuitem.GetMenuItemCount()):
                item = menuitem.FindItemByPosition(m)
                stritem = item.GetItemLabelText()
                stritem = stritem.split('\t')[0]
                if stritem == pathlist[-1].split('\t')[0] and item.IsSubMenu():
                    menuitem.DestroyItem(item)
                    return True
        else:
            item = menuitem.FindItemById(id)
            if item is None:
                return False
            menuitem.DestroyItem(item)
            # unbind the event and delete from menuAddon list
            self.Unbind(wx.EVT_MENU, id=id)
            if self.menuAddon[id][1]:
                self.Unbind(wx.EVT_UPDATE_UI, id=id)
            del self.menuAddon[id]

            return True
        return False

    def OnMenuAddOn(self, event):
        idx = event.GetId()
        signal = self.menuAddon.get(idx, None)
        if signal:
            signal = signal[0]
            dp.send(signal=signal, command=idx)

    def OnMenuCmdUI(self, event):
        idx = event.GetId()
        signal = self.menuAddon.get(idx, None)
        if signal:
            signal = signal[1]
            dp.send(signal=signal, event=event)
        else:
            event.Enable(True)

    def AddPanel(self,
                 panel,
                 title='Untitle',
                 active=True,
                 paneInfo=None,
                 target=None,
                 showhidemenu=None,
                 icon=None,
                 maximize=False,
                 direction='top',
                 minsize=None):
        """add the panel to AUI"""
        if not panel:
            return False
        panel.Reparent(self)
        # hide the window for now to avoid flicker
        panel.Hide()
        # if the target is None, find the notebook control that has the same
        # type as panel. It tries to put the same type panels in the same
        # notebook
        if target is None:
            for pane in self._mgr.GetAllPanes():
                if isinstance(pane.window, type(panel)):
                    target = pane.window
                    break
        elif isinstance(target, six.string_types):
            # find the target panel with caption
            for pane in self._mgr.GetAllPanes():
                if pane.caption == target:
                    target = pane.window
                    break
        targetpane = None
        try:
            if target:
                targetpane = self._mgr.GetPane(target)
                if targetpane and not targetpane.IsOk():
                    targetpane = None
        except:
            targetpane = None

        auipaneinfo = paneInfo
        dirs = {
            'top': aui.AUI_DOCK_TOP,
            'bottom': aui.AUI_DOCK_BOTTOM,
            'left': aui.AUI_DOCK_LEFT,
            'right': aui.AUI_DOCK_RIGHT,
            'center': aui.AUI_DOCK_CENTER
        }
        direction = dirs.get(direction, aui.AUI_DOCK_TOP)
        if auipaneinfo is None:
            # default panel settings. dock_row = -1 to add the pane to the
            # dock with same direction and layer, and dock_pos = 99 (a large
            # number) to add it to the right side
            auipaneinfo = aui.AuiPaneInfo().BestSize((300, 300)).Snappable()\
                          .Dockable().MinimizeButton(True).MaximizeButton(True)\
                          .Row(-1).Position(99)

        auipaneinfo.Caption(title).DestroyOnClose(not showhidemenu).Icon(icon)\
                   .Direction(direction)

        if not self._mgr.GetAllPanes():
            # set the first pane to be center pane
            auipaneinfo.CenterPane()
            active = True

        # auto generate the unique panel name
        name = "pane-%d" % self._pane_num
        self._pane_num += 1
        auipaneinfo.Name(name)
        if minsize:
            auipaneinfo.MinSize(minsize)

        panel.SetLabel(title)
        self._mgr.AddPane(panel, auipaneinfo, target=targetpane)
        if maximize:
            self._mgr.MaximizePane(auipaneinfo)
        if targetpane:
            self.ShowPanel(targetpane.window, targetpane.IsShown())
        self.ShowPanel(panel, active)
        # add the menu item to show/hide the panel
        if showhidemenu:
            mid = self.AddMenu(showhidemenu,
                               rxsignal='frame.check_menu',
                               updatesignal='frame.update_menu',
                               kind='Check',
                               autocreate=True)
            if mid != wx.NOT_FOUND:
                self.paneAddon[mid] = {'panel': panel, 'path': showhidemenu}
        return True

    def DeletePanel(self, panel):
        """hide and destroy the panel"""
        # delete the show/hide menu
        for (pid, pane) in six.iteritems(self.paneAddon):
            if panel == pane['panel']:
                self.DeleteMenu(pane['path'], pid)
                del self.paneAddon[pid]
                break

        # delete the panel from the manager
        pane = self._mgr.GetPane(panel)
        if pane is None or not pane.IsOk():
            return False
        pane.DestroyOnClose(True)
        self._mgr.ClosePane(pane)
        self._mgr.Update()
        return True

    def ShowPanel(self, panel, show=True, focus=False):
        """show/hide the panel"""
        pane = self._mgr.GetPane(panel)
        if pane is None or not pane.IsOk():
            return False
        self._mgr.ShowPane(panel, show)
        if focus:
            panel.SetFocus()
        return True

    def TogglePanel(self, command):
        """toggle the display of the panel"""
        pane = self.paneAddon.get(command, None)
        if not pane:
            return
        panel = pane['panel']
        # IsShown may not work, since the panel may be hidden while IsShown()
        # returns True
        show = panel.IsShownOnScreen()
        self.ShowPanel(panel, not show)

    def UpdateMenu(self, event):
        """update the menu checkbox"""
        pane = self.paneAddon.get(event.GetId(), None)
        if not pane:
            return
        panel = pane['panel']
        show = panel.IsShownOnScreen()
        event.Check(show)



#    BaseValidator.py
#
#    ------------------------------------------------------------
#    Copyright 2004 by Samuel Reynolds. All rights reserved.
#
#    Permission to use, copy, modify, and distribute this software and its
#    documentation for any purpose and without fee is hereby granted,
#    provided that the above copyright notice appear in all copies and that
#    both that copyright notice and this permission notice appear in
#    supporting documentation, and that the name of Samuel Reynolds
#    not be used in advertising or publicity pertaining to distribution
#    of the software without specific, written prior permission.
#
#    SAMUEL REYNOLDS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
#    INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO
#    EVENT SHALL SAMUEL REYNOLDS BE LIABLE FOR ANY SPECIAL, INDIRECT, OR
#    CONSEQUENTIAL DAMAGES, OR FOR ANY DAMAGES WHATSOEVER RESULTING FROM
#    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
#    NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
#    WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#    ------------------------------------------------------------

import sys
import traceback
import wx


class BaseValidator(wx.Validator):
    """
    Base class for validators that validate *attributes* of
    objects and provide two-way data transfer between those
    attributes and UI controls.

    BaseValidator has no knowledge of the specific
    type of widget to be validated. Responsibility for
    interacting with specific types of widgets is delegated
    to subclasses.

    Subclasses must implement the following methods:
        * _setControlValue(self, value)
                Set the value of control to the value provided.
        * _getControlValue(self)
                Return the value of control.
    Remaining logic is implemented in BaseValidator.


    CHAMELEON POWERS:

    To simplify use in an object-editor context (in which
    a single editor GUI is used to edit multiple objects
    of the same type), BaseValidator allows changing
    objects on the fly.

    For example, if references to field widgets are stored
    in a list self._fieldWidgets, the editor class could
    implement the following methods:

        def SetObject(self, editObject):
            "Set object for editing."
            self._editObject = editObject
            self._updateValidators()
            self.TransferDataToWindow()

        def _updateValidators(self):
            for wgt in self._fieldWidgets:
                validator = wgt.GetValidator()
                validator.SetObject(self._editObject)


    FORMATTERS

    BaseValidator uses Formatters for validation and
    two-way reformatting. A Formatter is aware of source data type,
    and handles translation between presentation representation and
    storage representation.

    A Formatter must provide the following interface methods:
        * format(stored_value)
                Format a value for presentation.
        * coerce(presentation_value)
                Convert a presentation-formatted value
                to a storage-formatted value (may include
                type conversion such as string->integer).
        * validate(presentation_value)
                Validate a presentation-formatted value.
                Return True if valid or False if invalid.

    If a formatter is assigned, the BaseValidator
    will call the formatter's format, coerce, and validate
    methods at appropriate points in its processing.

    With a formatter, there is no need of, for example,
    integer-aware entry fields. The formatter handles
    validation and conversion of the input value.
    """
    def __init__(self,
                 obj,
                 attr,
                 formatter=None,
                 required=True,
                 callback=None):
        super(BaseValidator, self).__init__()

        self.obj = obj
        self.attr = attr
        self.required = required
        self.formatter = formatter
        self.callback = callback

    def Clone(self):
        """
        Return a new validator for the same field of the same object.
        """
        return self.__class__(self.obj, self.attr, self.formatter,
                              self.required, self.callback)

    def SetObject(self, obj):
        """
        Set or change the object with which the validator interacts.

        Useful for changing objects when a single panel is used
        to edit a number of identical objects.
        """
        self.obj = obj

    def TransferToWindow(self):
        """
        Transfer data from validator to window.

        Delegates actual writing to destination widget to subclass
        via _setControlValue method.
        """
        try:
            if self.obj is None:
                # Nothing to do
                return True

            # Copy object attribute value to widget
            val = getattr(self.obj, self.attr)
            if val is None:
                val = ''
            if self.formatter:
                val = self.formatter.format(val)
            self._setControlValue(val)

            return True
        except:
            traceback.print_exc(file=sys.stdout)

        return False

    def TransferFromWindow(self):
        """
        Transfer data from window to validator.

        Delegates actual reading from destination widget to subclass
        via _getControlValue method.

        Only copies data if value is actually changed from attribute value.
        """
        try:
            if self.obj is None:
                # Nothing to do
                return True

            # Get widget value
            val = self._getControlValue()

            # Check widget value against attribute value; only copy if changed
            # Get object attribute value
            old_val = getattr(self.obj, self.attr)
            if self.formatter:
                old_val = self.formatter.format(old_val)
            if val != old_val:
                if self.formatter:
                    val = self.formatter.coerce(val)
                setattr(self.obj, self.attr, val)

            return True
        except:
            traceback.print_exc(file=sys.stdout)

        return False

    def Validate(self, win):
        """
        Validate the contents of the given control.

        Default behavior: Anything goes.
        """
        valid = True
        try:
            val = self._getControlValue()
            if self.required and val == '':
                valid = False
            if valid and self.formatter:
                valid = self.formatter.validate(val)
            if self.callback:
                self.callback(self.obj, self.attr, val, self.required, valid)
        except:
            traceback.print_exc(file=sys.stdout)

        return valid

    def _setControlValue(self, value):
        """
        Set the value of the target control.

        Subclass must implement.
        """
        raise NotImplementedError('Subclass must implement _setControlValue')

    def _getControlValue(self):
        """
        Return the value from the target control.

        Subclass must implement.
        """
        raise NotImplementedError('Subclass must implement _getControlValue')


class TextValidator(BaseValidator):
    """
    Validator for TextCtrl widgets.
    """
    def __init__(self, *args, **kwargs):
        """ Standard constructor. """
        super(TextValidator, self).__init__(*args, **kwargs)

        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)

    def OnKeyDown(self, event):
        event.Skip()
        textCtrl = self.GetWindow()
        text = textCtrl.GetValue()
        wx.CallAfter(self.Validate, textCtrl)

    def Validate(self, win):
        valid = super(TextValidator, self).Validate(win)
        try:
            if not valid:
                win.SetForegroundColour("red")
                win.Refresh()
            else:
                win.SetForegroundColour(wx.NullColour)
                win.Refresh()
        except:
            traceback.print_exc(file=sys.stdout)

        return valid

    def TransferToWindow(self):
        """
        Transfer data from validator to window.

        Delegates actual writing to destination widget to subclass
        via _setControlValue method.
        """
        if self.obj is None:
            # Clear the widget
            wgt = self.GetWindow()
            wgt.Clear()
            return True

        # Default behavior otherwise
        return super(TextValidator, self).TransferToWindow()

    def _setControlValue(self, value):
        """
        Set the value of the TextCtrl.
        """
        wgt = self.GetWindow()
        wgt.SetValue(value)

    def _getControlValue(self):
        """
        Return the value from the TextCtrl.
        """
        wgt = self.GetWindow()
        return wgt.GetValue()


class SelectorValidator(BaseValidator):
    """
    Validator for ControlWithItems widgets (ListBox, Choice).

    For wx.ListBox, assumes single-selection mode (wx.LB_SINGLE).
    """
    def __init__(self, obj, attr, formatter, *args, **kwargs):
        """ Standard constructor. """
        super(SelectorValidator, self).__init__(obj, attr, formatter, *args,
                                                **kwargs)

    def _getFieldOptions(self, name):
        """
        Return list of (id,label) pairs.
        """
        return self.formatter.validValues()

    def _setControlValue(self, value):
        """
        Set the value *and the options* of the control.
        By the time this is called, the value is already mapped for display.
        """
        wgt = self.GetWindow()
        # Get options (list of (id,value) pairs)
        options = self._getFieldOptions(self.attr)
        # Sort alphabetically
        options = [(opt[1], opt) for opt in options]
        #options.sort()
        options = [opt[1] for opt in options]
        # Replace selector contents
        wgt.Clear()
        for id, label in options:
            wgt.Append(label, id)

        # Set selection
        wgt.SetStringSelection(value)

    def _getControlValue(self):
        """
        Return the value from the TextCtrl.
        """
        wgt = self.GetWindow()
        return wgt.GetStringSelection()


class CheckListBoxValidator(BaseValidator):
    """
    Validator for CheckListBox widgets.
    """
    def __init__(self, obj, attr, formatter, *args, **kwargs):
        """ Standard constructor. """
        super(CheckListBoxValidator, self).__init__(obj, attr, formatter,
                                                    *args, **kwargs)

    ########## REQUIRED INTERFACE ##########

    def _getFieldOptions(self, name):
        """
        Return list of (id,label) pairs.
        """
        return self.formatter.validValues()

    def _setControlValue(self, value):
        """
        Set the value *and the options* of the control.
        By the time this is called, the value is already mapped for display.

        @param value:    Sequence of (value, label) pairs.
        """
        wgt = self.GetWindow()

        # Get options (list of (id,value) pairs)
        options = self._getFieldOptions(self.attr)
        # Sort alphabetically
        options = [(opt[1], opt) for opt in options]
        #options.sort()
        options = [opt[1] for opt in options]
        # Replace selector contents
        self._setControlOptions(options)

        # Set selection
        wgt._setControlSelections(value)

    def _getControlValue(self):
        """
        Return the value from the TextCtrl.

        Returns a list of client data values (not row indices).
        """
        wgt = self.GetWindow()
        selections = wgt.GetStringSelections()
        value = [wgt.GetClientData(idx) for idx in selections]
        return value

    ########## END REQUIRED INTERFACE ##########

    def _setControlOptions(self, options):
        """
        Set up or update control options.

        @param options:    Sequence of (id,label) pairs.
        """
        wgt = self.GetWindow()
        wgt.Clear()
        for id, label in options:
            wgt.Append(label, id)

    def _setControlSelections(self, value):
        """
        Select the specified items in the control, and unselect others.

        @param value:    Integer or sequence of integers representing
                        the data value(s) of item(s) to be selected.

                        If None or empty sequence, all currently-selected
                        items will be deselected.

                        Any items in value that are not found in the
                        control have no effect.
        """
        if value is None:
            value = tuple()
        elif not isinstance(value, (list, tuple)):
            value = (value, )

        wgt = self.GetWindow()
        for idx in range(0, wgt.GetCount()):
            item_data = wgt.GetClientData(idx)
            if item_data in value:
                if not wgt.IsChecked(idx):
                    wgt.Check(idx, True)
            else:
                if wgt.IsChecked(idx):
                    wgt.Check(idx, False)


class RadioBoxValidator(BaseValidator):
    """
    Validator for RadioBox widgets.
    """
    def __init__(self, obj, attr, formatter, *args, **kwargs):
        """ Standard constructor. """
        super(RadioBoxValidator, self).__init__(obj, attr, formatter, *args,
                                                **kwargs)

    def _setControlValue(self, value):
        """
        Set the value *and the options* of the control.
        By the time this is called, the value is already mapped for display.
        """
        wgt = self.GetWindow()
        wgt.SetSelection(wgt.FindString(value))

    def _getControlValue(self):
        """
        Return the value from the TextCtrl.
        """
        wgt = self.GetWindow()
        return wgt.GetString(wgt.GetSelection())


class SpinSliderValidator(BaseValidator):
    """
    Validator for wx.SpinCtrl and wx.Slider widgets.

    """
    def __init__(self, obj, attr, formatter, *args, **kwargs):
        """ Standard constructor. """
        super(SpinSliderValidator, self).__init__(obj, attr, formatter, *args,
                                                  **kwargs)

    def _setControlValue(self, value):
        """
        Set the value *and the options* of the control.
        By the time this is called, the value is already mapped for display.
        """
        wgt = self.GetWindow()
        # Get options (list of (id,value) pairs)
        max_val = 100
        min_val = 0
        if hasattr(self.formatter, 'max_val'):
            max_val = self.formatter.max_val
        if hasattr(self.formatter, 'min_val'):
            min_val = self.formatter.min_val
        wgt.SetMax(max_val)
        wgt.SetMin(min_val)
        wgt.SetValue(self.formatter.coerce(value))

    def _getControlValue(self):
        """
        Return the value from the TextCtrl.
        """
        wgt = self.GetWindow()
        return wgt.GetValue()



"""Subclass of mainFrame, which is generated by wxFormBuilder."""

import sys
import importlib
import traceback
import json
import six
import six.moves
import wx
import wx.lib.agw.aui as aui
import wx.py
import wx.py.dispatcher as dp
import wx.adv
from .frameplus import FramePlus
from .mainframexpm import bsmedit_xpm, header_xpm, bsmedit_svg
from . import __version__, to_byte
from .bsm.utility import PopupMenu, svg_to_bitmap
from .bsm import auto_load_module

class FileDropTarget(wx.FileDropTarget):
    def __init__(self):
        wx.FileDropTarget.__init__(self)

    def OnDropFiles(self, x, y, filenames):
        for fname in filenames:
            wx.CallAfter(dp.send, signal='frame.file_drop', filename=fname)
        return True

class TaskBarIcon(wx.adv.TaskBarIcon):
    TBMENU_RESTORE = wx.NewId()
    TBMENU_CLOSE   = wx.NewId()
    TBMENU_CHANGE  = wx.NewId()
    TBMENU_REMOVE  = wx.NewId()

    def __init__(self, frame, icon):
        wx.adv.TaskBarIcon.__init__(self, iconType=wx.adv.TBI_DOCK)
        self.frame = frame

        # Set the image
        self.SetIcon(icon, "bsmedit")
        self.imgidx = 1

        # bind some events
        #self.Bind(wx.EVT_TASKBAR_LEFT_DCLICK, self.OnTaskBarActivate)
        self.Bind(wx.EVT_MENU, self.OnTaskBarActivate, id=self.TBMENU_RESTORE)
        self.Bind(wx.EVT_MENU, self.OnTaskBarClose, id=self.TBMENU_CLOSE)


    def CreatePopupMenu(self):
        """
        This method is called by the base class when it needs to popup
        the menu for the default EVT_RIGHT_DOWN event.  Just create
        the menu how you want it and return it from this function,
        the base class takes care of the rest.
        """
        menu = wx.Menu()
        menu.Append(self.TBMENU_RESTORE, "Restore bsmedit")
        menu.Append(self.TBMENU_CLOSE,   "Close bsmedit")
        return menu


    def MakeIcon(self, img):
        """
        The various platforms have different requirements for the
        icon size...
        """
        if "wxMSW" in wx.PlatformInfo:
            img = img.Scale(16, 16)
        elif "wxGTK" in wx.PlatformInfo:
            img = img.Scale(22, 22)
        # wxMac can be any size upto 128x128, so leave the source img alone....
        icon = wx.Icon(img.ConvertToBitmap() )
        return icon


    def OnTaskBarActivate(self, evt):
        if self.frame.IsIconized():
            self.frame.Iconize(False)
        if not self.frame.IsShown():
            self.frame.Show(True)
        self.frame.Raise()


    def OnTaskBarClose(self, evt):
        wx.CallAfter(self.frame.Close)


class MainFrame(FramePlus):

    ID_VM_RENAME = wx.NewId()
    ID_CONTACT = wx.NewId()

    def __init__(self, parent, **kwargs):
        FramePlus.__init__(self,
                           parent,
                           title='bsmedit',
                           size=wx.Size(800, 600),
                           style=wx.DEFAULT_FRAME_STYLE | wx.TAB_TRAVERSAL)
        self.InitMenu()
        self._mgr.SetAGWFlags(self._mgr.GetAGWFlags()
                              | aui.AUI_MGR_ALLOW_ACTIVE_PANE
                              | aui.AUI_MGR_SMOOTH_DOCKING
                              | aui.AUI_MGR_USE_NATIVE_MINIFRAMES
                              | aui.AUI_MGR_LIVE_RESIZE)
        # set mainframe icon
        icon = wx.Icon()
        icon.CopyFromBitmap(svg_to_bitmap(bsmedit_svg))
        self.SetIcon(icon)

        if 'wxMac' in wx.PlatformInfo:
            icon.CopyFromBitmap(svg_to_bitmap(bsmedit_svg, 1024, 1024))
            self.tbicon = TaskBarIcon(self, icon)

        # status bar
        self.statusbar = wx.StatusBar(self)
        self.SetStatusBar(self.statusbar)
        self.statusbar_width = [-1]
        self.statusbar.SetStatusWidths(self.statusbar_width)

        # persistent configuration
        conf = kwargs.get('config', 'bsmedit')
        self.config = wx.FileConfig(conf, style=wx.CONFIG_USE_LOCAL_FILE)

        # recent file list
        self.filehistory = wx.FileHistory(8)
        self.config.SetPath('/FileHistory')
        self.filehistory.Load(self.config)
        self.filehistory.UseMenu(self.menuRecentFiles)
        self.filehistory.AddFilesToMenu()
        self.Bind(wx.EVT_MENU_RANGE,
                  self.OnMenuFileHistory,
                  id=wx.ID_FILE1,
                  id2=wx.ID_FILE9)

        self.closing = False

        # Create & Link the Drop Target Object to main window
        self.SetDropTarget(FileDropTarget())

        self.Bind(wx.EVT_ACTIVATE, self.OnActivate)
        self.Bind(aui.EVT_AUI_PANE_ACTIVATED, self.OnPaneActivated)
        dp.connect(self.SetPanelTitle, 'frame.set_panel_title')
        dp.connect(self.ShowStatusText, 'frame.show_status_text')
        dp.connect(self.AddFileHistory, 'frame.add_file_history')
        dp.connect(self.SetConfig, 'frame.set_config')
        dp.connect(self.GetConfig, 'frame.get_config')

        # append sys path
        sys.path.append('.')
        for p in kwargs.get('path', []):
            sys.path.append(p)

        self.bsm_packages = auto_load_module
        self.addon = {}
        self.InitAddOn(kwargs.get('module', ()))

        # initialization done, broadcasting the message so plugins can do some
        # after initialization processing.
        dp.send('frame.initialized')

        # load the perspective
        if not kwargs.get('ignore_perspective', False):
            perspective = self.GetConfig('mainframe', 'perspective')
            if perspective and not wx.GetKeyState(wx.WXK_SHIFT):
                self._mgr.LoadPerspective(perspective)
                self.UpdatePaneMenuLabel()

        self.Bind(aui.EVT_AUINOTEBOOK_TAB_RIGHT_DOWN, self.OnPageRightDown)
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)

    def InitMenu(self):
        """initialize the menubar"""
        menubar = wx.MenuBar()
        self.SetMenuBar(menubar)

        self.AddMenu('&File:New', kind="Popup", autocreate=True)
        self.AddMenu('&File:Open', kind="Popup")
        self.AddMenu('&File:Sep', kind="Separator")
        self.AddMenu('&File:Recent Files', kind="Popup")
        self.menuRecentFiles = self.GetMenu(['File', 'Recent Files'])
        self.AddMenu('&File:Sep', kind="Separator")
        self.AddMenu('&File:&Quit', id=wx.ID_CLOSE)

        # add the common Edit menus to menubar; otherwise, the context-menu
        # from shell or editor may not work (e.g., Mac)
        self.AddMenu('&Edit:&Undo', id=wx.ID_UNDO, autocreate=True)
        self.AddMenu('&Edit:&Redo', id=wx.ID_REDO)
        self.AddMenu('&Edit:Sep', kind="Separator")
        self.AddMenu('&Edit:&Cut', id=wx.ID_CUT)
        self.AddMenu('&Edit:&Copy', id=wx.ID_COPY)
        self.AddMenu('&Edit:&Paste', id=wx.ID_PASTE)
        self.AddMenu('&Edit:&Clear', id=wx.ID_CLEAR)
        self.AddMenu('&Edit:Sep', kind="Separator")
        self.AddMenu('&Edit:&Select All', id=wx.ID_SELECTALL)

        self.AddMenu('&View:Toolbars', kind="Popup", autocreate=True)
        self.AddMenu('&View:Sep', kind="Separator")
        self.AddMenu('&View:Panels', kind="Popup")

        self.AddMenu('&Tools', kind="Popup", autocreate=True)

        self.AddMenu('&Help:&Home', id=wx.ID_HOME, autocreate=True)
        self.ID_CONTACT = self.AddMenu('&Help:&Contact')
        self.AddMenu('&Help:Sep', kind="Separator")
        self.AddMenu('&Help:About', id=wx.ID_ABOUT)

        # Connect Events
        self.Bind(wx.EVT_MENU, self.OnFileQuit, id=wx.ID_CLOSE)
        self.Bind(wx.EVT_MENU, self.OnHelpHome, id=wx.ID_HOME)
        self.Bind(wx.EVT_MENU, self.OnHelpContact, id=self.ID_CONTACT)
        self.Bind(wx.EVT_MENU, self.OnHelpAbout, id=wx.ID_ABOUT)

    def InitAddOn(self, modules):
        if not modules:
            # load all modules
            modules = ["default"]

        for module in modules:
            module = module.split('+')
            options = {}
            if len(module) == 2:
                if all([c in 'htblr' for c in module[1]]):
                    if 'h' in module[1]:
                        options['active'] = False
                    if 't' in module[1]:
                        options['direction'] = 'Top'
                    if 'b' in module[1]:
                        options['direction'] = 'bottom'
                    if 'l' in module[1]:
                        options['direction'] = 'left'
                    if 'r' in module[1]:
                        options['direction'] = 'right'
                options['data'] = module[1]
            module = module[0]
            if module == 'default':
                module = self.bsm_packages
            else:
                module = [module]
            for pkg in module:
                if pkg in self.bsm_packages:
                    # module in bsm
                    pkg = 'bsmedit.bsm.%s' % pkg

                if pkg in self.addon:
                    # already loaded
                    continue
                self.addon[pkg] = False
                try:
                    mod = importlib.import_module(pkg)
                except ImportError:
                    traceback.print_exc(file=sys.stdout)
                else:
                    if hasattr(mod, 'bsm_initialize'):
                        mod.bsm_initialize(self, **options)
                        self.addon[pkg] = True
                    else:
                        print("Error: Invalid module: %s" % pkg)

    def AddFileHistory(self, filename):
        """add the file to recent file list"""
        self.config.SetPath('/FileHistory')
        self.filehistory.AddFileToHistory(filename)
        self.filehistory.Save(self.config)
        self.config.Flush()

    def SetConfig(self, group, **kwargs):
        if not group.startswith('/'):
            group = '/' + group
        for key, value in six.iteritems(kwargs):
            if key in ['signal', 'sender']:
                # reserved key for dp.send
                continue
            if not isinstance(value, str):
                # add sign to indicate that the value needs to be deserialize
                value = '__bsm__' + json.dumps(value)
            self.config.SetPath(group)
            self.config.Write(key, value)

    def GetConfig(self, group, key):
        if not group.startswith('/'):
            group = '/' + group
        if self.config.HasGroup(group):
            self.config.SetPath(group)
            if self.config.HasEntry(key):
                value = self.config.Read(key)
                if value.startswith('__bsm__'):
                    value = json.loads(value[7:])
                return value
        return None

    def OnPageRightDown(self, evt):
        # get the index inside the current tab control
        idx = evt.GetSelection()
        tabctrl = evt.GetEventObject()
        tabctrl.SetSelection(idx)
        page = tabctrl.GetPage(idx)
        self.RenamePanel(page)

    def OnRightDown(self, evt):
        evt.Skip()

        part = self._mgr.HitTest(*evt.GetPosition())
        if not part or part.pane.IsNotebookControl():
            return

        self.RenamePanel(part.pane.window)

    def RenamePanel(self, panel):
        if not panel:
            return

        menu = wx.Menu()
        menu.Append(self.ID_VM_RENAME, "&Rename")
        command = PopupMenu(self, menu)

        if command == self.ID_VM_RENAME:
            pane = self._mgr.GetPane(panel)
            if not pane:
                return
            name = pane.caption
            name = wx.GetTextFromUser("Type in the name:", "Input Name", name,
                                      self)
            # when user click 'cancel', name will be empty, ignore it.
            if name and name != pane.caption:
                self.SetPanelTitle(pane.window, name)

    def UpdatePaneMenuLabel(self):
        # update the menu
        for (pid, panel) in six.iteritems(self.paneAddon):
            pathlist = panel['path'].split(':')
            menuitem = self.GetMenu(pathlist[:-1])
            if not menuitem:
                continue
            pane = self._mgr.GetPane(panel['panel'])
            item = menuitem.FindItemById(pid)
            if item and pane.caption != item.GetItemLabelText():
                item.SetItemLabel(pane.caption)

    def OnCloseWindow(self, evt):
        self.tbicon.Destroy()
        evt.Skip()

    def OnClose(self, event):
        """close the main program"""
        dp.send('frame.closing', event=event)
        if event.GetVeto():
            return
        self.closing = True
        self.SetConfig('mainframe', perspective=self._mgr.SavePerspective())
        dp.send('frame.exit')
        self.config.Flush()
        super(MainFrame, self).OnClose(event)

    def ShowStatusText(self, text, index=0, width=-1):
        """set the status text"""
        if index >= len(self.statusbar_width):
            exd = [0] * (index + 1 - len(self.statusbar_width))
            self.statusbar_width.extend(exd)
            self.statusbar.SetFieldsCount(index + 1)
        if self.statusbar_width[index] < width:
            self.statusbar_width[index] = width
            self.statusbar.SetStatusWidths(self.statusbar_width)
        self.statusbar.SetStatusText(text, index)

    def OnActivate(self, event):
        if not self.closing:
            dp.send('frame.activate', activate=event.GetActive())
        event.Skip()

    def OnPaneActivated(self, event):
        """notify the window managers that the panel is activated"""
        if self.closing:
            return
        pane = event.GetPane()
        if isinstance(pane, aui.auibook.AuiNotebook):
            window = pane.GetCurrentPage()
        else:
            window = pane

        dp.send('frame.activate_panel', pane=window)

    def SetPanelTitle(self, pane, title):
        """set the panel title"""
        if pane:
            info = self._mgr.GetPane(pane)
            if info and info.IsOk() and info.caption != title:
                info.Caption(title)
                self._mgr.RefreshPaneCaption(pane)
                self.UpdatePaneMenuLabel()

    # Handlers for mainFrame events.
    def OnFileQuit(self, event):
        """close the program"""
        self.Close(True)

    def OnHelpHome(self, event):
        """go to homepage"""
        wx.BeginBusyCursor()
        import webbrowser
        webbrowser.open("http://bsmedit.feiyilin.com")
        wx.EndBusyCursor()

    def OnHelpContact(self, event):
        """send email"""
        wx.BeginBusyCursor()
        import webbrowser
        webbrowser.open("mail:tq@feiyilin.com")
        wx.EndBusyCursor()

    def OnHelpAbout(self, event):
        """show about dialog"""
        dlg = AboutDialog(self)
        dlg.ShowModal()
        dlg.Destroy()

    def OnMenuFileHistory(self, event):
        """open the recent file"""
        fileNum = event.GetId() - wx.ID_FILE1
        path = self.filehistory.GetHistoryFile(fileNum)
        self.filehistory.AddFileToHistory(path)
        dp.send('frame.file_drop', filename=path)


class AboutDialog(wx.Dialog):
    def __init__(self, parent):
        wx.Dialog.__init__(self,
                           parent,
                           title="About bsmedit",
                           style=wx.DEFAULT_DIALOG_STYLE)

        szAll = wx.BoxSizer(wx.VERTICAL)

        self.panel = wx.Panel(self, style=wx.TAB_TRAVERSAL)
        self.panel.SetBackgroundColour(wx.WHITE)

        szPanel = wx.BoxSizer(wx.VERTICAL)

        self.header = wx.StaticBitmap(self.panel)
        self.header.SetBitmap(wx.Bitmap(to_byte(header_xpm)))
        szPanel.Add(self.header, 0, wx.EXPAND, 0)

        caption = 'bsmedit %s' % (__version__)
        self.stCaption = wx.StaticText(self.panel, wx.ID_ANY, caption)
        self.stCaption.SetFont(wx.Font(16, 74, 90, 92, False, "Arial"))

        szPanel.Add(self.stCaption, 0, wx.ALL | wx.EXPAND, 5)

        strCopyright = '(c) 2018 Tianzhu Qiao. All rights reserved.'

        self.stCopyright = wx.StaticText(self.panel, wx.ID_ANY, strCopyright)
        self.stCopyright.SetMaxSize((240, -1))
        self.stCopyright.SetFont(wx.Font(8, 74, 90, 90, False, "Arial"))
        szPanel.Add(self.stCopyright, 0, wx.ALL | wx.EXPAND, 5)

        build = wx.GetOsDescription() + '; wxWidgets ' + wx.version()
        self.stBuild = wx.StaticText(self.panel, wx.ID_ANY, build)
        self.stBuild.SetMaxSize((240, -1))
        self.stBuild.Wrap(240)
        self.stBuild.SetFont(wx.Font(8, 74, 90, 90, False, "Arial"))
        szPanel.Add(self.stBuild, 0, wx.ALL | wx.EXPAND, 5)

        stLine = wx.StaticLine(self.panel, style=wx.LI_HORIZONTAL)
        szPanel.Add(stLine, 1, wx.EXPAND | wx.ALL, 0)

        self.panel.SetSizer(szPanel)
        self.panel.Layout()
        szPanel.Fit(self.panel)

        szAll.Add(self.panel, 1, wx.EXPAND | wx.ALL, 0)

        szConfirm = wx.BoxSizer(wx.VERTICAL)
        self.btnOk = wx.Button(self, wx.ID_OK, u"Ok")
        szConfirm.Add(self.btnOk, 0, wx.ALIGN_RIGHT | wx.ALL, 5)

        szAll.Add(szConfirm, 0, wx.EXPAND, 5)

        self.SetSizer(szAll)
        self.Layout()
        szAll.Fit(self)



# -*- coding: utf-8 -*-
#
# Sorted set implementation.

from .sortedlist import SortedList, recursive_repr, SortedListWithKey
from collections import Set, MutableSet, Sequence
from itertools import chain
import operator as op

class SortedSet(MutableSet, Sequence):
    """
    A `SortedSet` provides the same methods as a `set`.  Additionally, a
    `SortedSet` maintains its items in sorted order, allowing the `SortedSet` to
    be indexed.

    Unlike a `set`, a `SortedSet` requires items be hashable and comparable.
    """
    def __init__(self, iterable=None, key=None, load=1000, _set=None):
        """
        A `SortedSet` provides the same methods as a `set`.  Additionally, a
        `SortedSet` maintains its items in sorted order, allowing the
        `SortedSet` to be indexed.

        An optional *iterable* provides an initial series of items to populate
        the `SortedSet`.

        An optional *key* argument defines a callable that, like the `key`
        argument to Python's `sorted` function, extracts a comparison key from
        each set item. If no function is specified, the default compares the
        set items directly.

        An optional *load* specifies the load-factor of the set. The default
        load factor of '1000' works well for sets from tens to tens of millions
        of elements.  Good practice is to use a value that is the cube root of
        the set size.  With billions of elements, the best load factor depends
        on your usage.  It's best to leave the load factor at the default until
        you start benchmarking.
        """
        self._key = key
        self._load = load

        self._set = set() if _set is None else _set

        _set = self._set
        self.isdisjoint = _set.isdisjoint
        self.issubset = _set.issubset
        self.issuperset = _set.issuperset

        if key is None:
            self._list = SortedList(self._set, load=load)
        else:
            self._list = SortedListWithKey(self._set, key=key, load=load)

        _list = self._list
        self.bisect_left = _list.bisect_left
        self.bisect = _list.bisect
        self.bisect_right = _list.bisect_right
        self.index = _list.index
        self.irange = _list.irange
        self.islice = _list.islice

        if key is not None:
            self.bisect_key_left = _list.bisect_key_left
            self.bisect_key_right = _list.bisect_key_right
            self.bisect_key = _list.bisect_key
            self.irange_key = _list.irange_key

        if iterable is not None:
            self._update(iterable)

    def __contains__(self, value):
        """Return True if and only if *value* is an element in the set."""
        return (value in self._set)

    def __getitem__(self, index):
        """
        Return the element at position *index*.

        Supports slice notation and negative indexes.
        """
        return self._list[index]

    def __delitem__(self, index):
        """
        Remove the element at position *index*.

        Supports slice notation and negative indexes.
        """
        _list = self._list
        if isinstance(index, slice):
            values = _list[index]
            self._set.difference_update(values)
        else:
            value = _list[index]
            self._set.remove(value)
        del _list[index]

    def _make_cmp(set_op, doc):
        def comparer(self, that):
            if isinstance(that, SortedSet):
                return set_op(self._set, that._set)
            elif isinstance(that, Set):
                return set_op(self._set, that)
            else:
                return NotImplemented

        comparer.__name__ = '__{0}__'.format(set_op.__name__)
        doc_str = 'Return True if and only if Set is {0} `that`.'
        comparer.__doc__ = doc_str.format(doc)

        return comparer

    __eq__ = _make_cmp(op.eq, 'equal to')
    __ne__ = _make_cmp(op.ne, 'not equal to')
    __lt__ = _make_cmp(op.lt, 'a proper subset of')
    __gt__ = _make_cmp(op.gt, 'a proper superset of')
    __le__ = _make_cmp(op.le, 'a subset of')
    __ge__ = _make_cmp(op.ge, 'a superset of')

    def __len__(self):
        """Return the number of elements in the set."""
        return len(self._set)

    def __iter__(self):
        """
        Return an iterator over the Set. Elements are iterated in their sorted
        order.

        Iterating the Set while adding or deleting values may raise a
        `RuntimeError` or fail to iterate over all entries.
        """
        return iter(self._list)

    def __reversed__(self):
        """
        Return an iterator over the Set. Elements are iterated in their reverse
        sorted order.

        Iterating the Set while adding or deleting values may raise a
        `RuntimeError` or fail to iterate over all entries.
        """
        return reversed(self._list)

    def add(self, value):
        """Add the element *value* to the set."""
        if value not in self._set:
            self._set.add(value)
            self._list.add(value)

    def clear(self):
        """Remove all elements from the set."""
        self._set.clear()
        self._list.clear()

    def copy(self):
        """Create a shallow copy of the sorted set."""
        return self.__class__(key=self._key, load=self._load, _set=set(self._set))

    __copy__ = copy

    def count(self, value):
        """Return the number of occurrences of *value* in the set."""
        return 1 if value in self._set else 0

    def discard(self, value):
        """
        Remove the first occurrence of *value*.  If *value* is not a member,
        does nothing.
        """
        if value in self._set:
            self._set.remove(value)
            self._list.discard(value)

    def pop(self, index=-1):
        """
        Remove and return item at *index* (default last).  Raises IndexError if
        set is empty or index is out of range.  Negative indexes are supported,
        as for slice indices.
        """
        value = self._list.pop(index)
        self._set.remove(value)
        return value

    def remove(self, value):
        """
        Remove first occurrence of *value*.  Raises ValueError if
        *value* is not present.
        """
        self._set.remove(value)
        self._list.remove(value)

    def difference(self, *iterables):
        """
        Return a new set with elements in the set that are not in the
        *iterables*.
        """
        diff = self._set.difference(*iterables)
        new_set = self.__class__(key=self._key, load=self._load, _set=diff)
        return new_set

    __sub__ = difference
    __rsub__ = __sub__

    def difference_update(self, *iterables):
        """
        Update the set, removing elements found in keeping only elements
        found in any of the *iterables*.
        """
        values = set(chain(*iterables))
        if (4 * len(values)) > len(self):
            self._set.difference_update(values)
            self._list.clear()
            self._list.update(self._set)
        else:
            _discard = self.discard
            for value in values:
                _discard(value)
        return self

    __isub__ = difference_update

    def intersection(self, *iterables):
        """
        Return a new set with elements common to the set and all *iterables*.
        """
        comb = self._set.intersection(*iterables)
        new_set = self.__class__(key=self._key, load=self._load, _set=comb)
        return new_set

    __and__ = intersection
    __rand__ = __and__

    def intersection_update(self, *iterables):
        """
        Update the set, keeping only elements found in it and all *iterables*.
        """
        self._set.intersection_update(*iterables)
        self._list.clear()
        self._list.update(self._set)
        return self

    __iand__ = intersection_update

    def symmetric_difference(self, that):
        """
        Return a new set with elements in either *self* or *that* but not both.
        """
        diff = self._set.symmetric_difference(that)
        new_set = self.__class__(key=self._key, load=self._load, _set=diff)
        return new_set

    __xor__ = symmetric_difference
    __rxor__ = __xor__

    def symmetric_difference_update(self, that):
        """
        Update the set, keeping only elements found in either *self* or *that*,
        but not in both.
        """
        self._set.symmetric_difference_update(that)
        self._list.clear()
        self._list.update(self._set)
        return self

    __ixor__ = symmetric_difference_update

    def union(self, *iterables):
        """
        Return a new SortedSet with elements from the set and all *iterables*.
        """
        return self.__class__(chain(iter(self), *iterables), key=self._key, load=self._load)

    __or__ = union
    __ror__ = __or__

    def update(self, *iterables):
        """Update the set, adding elements from all *iterables*."""
        values = set(chain(*iterables))
        if (4 * len(values)) > len(self):
            self._set.update(values)
            self._list.clear()
            self._list.update(self._set)
        else:
            _add = self.add
            for value in values:
                _add(value)
        return self

    __ior__ = update
    _update = update

    def __reduce__(self):
        return (self.__class__, ((), self._key, self._load, self._set))

    @recursive_repr
    def __repr__(self):
        temp = '{0}({1}, key={2}, load={3})'
        return temp.format(
            self.__class__.__name__,
            repr(list(self)),
            repr(self._key),
            repr(self._load)
        )

    def _check(self):
        self._list._check()
        assert len(self._set) == len(self._list)
        _set = self._set
        assert all(val in _set for val in self._list)



import wx


class PropArtNative(object):
    def __init__(self):
        self.margin = {'top': 0, 'bottom': 0, 'left': 0, 'right': 0}
        self.gap_x = 2
        self.title_width = 150
        self.expansion_width = 12
        self.splitter_width = 8
        self.indent_width = 28
        self._font_label = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        self._font_value = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        self.text_clr = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNTEXT)
        self.text_clr_sel = wx.WHITE
        self.text_clr_disabled = wx.SystemSettings.GetColour(
            wx.SYS_COLOUR_GRAYTEXT)
        self.bg_clr = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW)
        self.bg_clr_sel = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        self.bg_clr_disabled = wx.SystemSettings.GetColour(
            wx.SYS_COLOUR_3DFACE)

    def SetTitleWidth(self, width):
        """set the title width"""
        self.title_width = width

    def GetTitleWidth(self):
        """return the width"""
        return self.title_width

    def SetLabelFont(self, font):
        """set label font"""
        self._font_label = font

    def GetLabelFont(self):
        """get label font"""
        return self._font_label

    def SetValueFont(self, font):
        """set value font"""
        self._font_value = font

    def GetValueFont(self):
        """get value font"""
        return self._font_value

    def SetTextColor(self, clr=None, clr_sel=None, clr_disabled=None):
        """
        set the text colors

        All values are string. If the value is None, the color will reset to
        default.
        """
        self.text_clr = clr
        if not self.text_clr:
            self.text_clr = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNTEXT)
        self.text_clr_sel = clr_sel
        if not self.text_clr_sel:
            self.text_clr_sel = wx.WHITE
        self.text_clr_disabled = clr_disabled
        if not self.text_clr_disabled:
            self.text_clr_disabled = wx.SystemSettings.GetColour(
                wx.SYS_COLOUR_GRAYTEXT)

    def GetTextColor(self):
        """get the text colors"""
        return (self.text_clr, self.text_clr_sel, self.text_clr_disabled)

    def SetBgColor(self, clr=None, clr_sel=None, clr_disabled=None):
        """
        set the background colors

        All values are string. If the value is None, the color will reset to
        default.
        """
        GetColour = wx.SystemSettings.GetColour
        self.bg_clr = clr
        if not self.bg_clr:
            self.bg_clr = GetColour(wx.SYS_COLOUR_WINDOW)
        self.bg_clr_sel = clr_sel
        if not self.bg_clr_sel:
            self.bg_clr_sel = GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        self.bg_clr_disabled = clr_disabled
        if not self.bg_clr_disabled:
            self.bg_clr_disabled = GetColour(wx.SYS_COLOUR_3DFACE)

    def GetBgColor(self):
        """get the background colors"""
        return (self.bg_clr, self.bg_clr_sel, self.bg_clr_disabled)

    def PrepareDrawRect(self, p):
        """calculate the rect for each section"""
        mx = self.gap_x
        irc = p.GetRect()
        irc.SetLeft(irc.GetLeft() + self.margin['left'])
        irc.SetRight(irc.GetRight() + self.margin['right'])
        irc.SetTop(irc.GetTop() + self.margin['top'])
        irc.SetBottom(irc.GetBottom() + self.margin['bottom'])
        x = irc.x
        x = x + mx * 2 + p.indent * self.indent_width

        if self.expansion_width > 0 and p.HasChildren():
            # expander icon
            rc = wx.Rect(*irc)
            rc.x = x + mx
            rc.SetWidth(self.expansion_width)
            p.regions['expander'] = rc
            x = rc.right

        # label
        p.regions['label'] = wx.Rect(*irc)
        p.regions['label'].x = x + mx * 2

        if not p.IsSeparator():
            title_width = p.title_width
            if title_width < 0:
                title_width = self.title_width
            p.regions['label'].SetRight(title_width)
            x = p.regions['label'].right

            rc = wx.Rect(*irc)
            rc.x = x + mx
            rc.SetWidth(self.splitter_width)
            p.regions['splitter'] = rc
            x = rc.right

            rc = wx.Rect(*irc)
            rc.x = x
            rc.SetWidth(irc.right - x)
            p.regions['value'] = rc
        else:
            # separator does not have splitter & value
            p.regions['label'].SetWidth(p.regions['label'].GetWidth() +
                                        irc.right - x)
            p.regions['splitter'] = wx.Rect(irc.right, irc.top, 0, 0)
            p.regions['value'] = wx.Rect(irc.right, irc.top, 0, 0)

    def DrawSplitter(self, dc, p):
        # draw splitter
        rcs = p.regions['splitter']
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DSHADOW)))
        dc.DrawLine(rcs.left, rcs.top, rcs.left, rcs.bottom)
        dc.DrawLine(rcs.right - 1, rcs.top, rcs.right - 1, rcs.bottom)
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DHILIGHT)))
        dc.DrawLine(rcs.left + 1, rcs.top, rcs.left + 1, rcs.bottom)
        dc.DrawLine(rcs.right, rcs.top, rcs.right, rcs.bottom)

    def DrawLabel(self, dc, p):
        # draw label
        if p.font_label is None:
            font = self._font_label
        else:
            font = wx.Font(p.font_label)
        dc.SetFont(font)

        if not p.IsEnabled() or p.IsReadonly():
            clr = wx.SystemSettings.GetColour(wx.SYS_COLOUR_GRAYTEXT)
        else:
            clr = wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNTEXT)
        dc.SetTextForeground(clr)
        rc = p.regions['label']
        dc.SetClippingRegion(rc)
        (w, h) = dc.GetTextExtent(p.label)

        dc.DrawText(p.label, rc.x, rc.y + (rc.height - h) / 2)
        p.show_label_tips = w > rc.width
        dc.DestroyClippingRegion()

    def DrawValue(self, dc, p):
        # draw value
        if p.font_value is None:
            font = self._font_value
        else:
            font = wx.Font(p.font_value)
        dc.SetFont(font)

        p.show_value_tips = False
        if p.window is None:
            crbg = p.bg_clr
            crtxt = wx.BLACK
            if not p.IsEnabled() or p.IsReadonly():
                crtxt = p.text_clr_disabled
                if not crtxt:
                    crtxt = self.text_clr_disabled
                crbg = p.bg_clr_disabled
                if not crbg:
                    crbg = self.bg_clr_disabled
            elif p.activated:
                crtxt = p.text_clr_sel
                if not crtxt:
                    crtxt = self.text_clr_sel
                crbg = p.bg_clr_sel
                if not crbg:
                    crbg = self.bg_clr_sel
            else:
                crtxt = p.text_clr
                if not crtxt:
                    crtxt = self.text_clr
                crbg = p.bg_clr
                if not crbg:
                    crbg = self.bg_clr

            dc.SetPen(wx.Pen(crtxt, 1, wx.PENSTYLE_TRANSPARENT))
            dc.SetBrush(wx.Brush(crbg))

            rc = p.regions['value']
            dc.DrawRectangle(rc)

            dc.SetTextForeground(crtxt)

            value = p.GetValueAsString()
            (w, h) = dc.GetTextExtent(value)
            dc.SetClippingRegion(rc)
            dc.DrawText(value, rc.x + 5, rc.top + (rc.height - h) / 2)
            p.show_value_tips = rc.width < w
            dc.DestroyClippingRegion()

    def DrawExpansion(self, dc, p):
        if self.expansion_width > 0 and p.HasChildren():
            w, h = self.expansion_width, self.expansion_width
            rc = p.regions['expander']
            x = rc.x + (rc.width - w) / 2
            y = rc.y + (rc.height - h) / 2 + 1
            dc.SetPen(wx.Pen(wx.BLACK))
            dc.SetBrush(wx.BLACK_BRUSH)
            render = wx.RendererNative.Get()
            if p.IsExpanded():
                render.DrawTreeItemButton(p.grid, dc, (x, y, w, h),
                                          wx.CONTROL_EXPANDED)
            else:
                render.DrawTreeItemButton(p.grid, dc, (x, y, w, h))

    def DrawBackground(self, dc, p):
        # draw background
        rc = p.GetRect()
        bg = p.GetGrid().GetBackgroundColour()
        pen = wx.Pen(wx.BLACK, 1, wx.PENSTYLE_TRANSPARENT)
        dc.SetPen(pen)
        brush = wx.Brush(bg)
        dc.SetBrush(brush)
        dc.DrawRectangle(rc.x, rc.y, rc.width, rc.height)

        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DSHADOW)))
        dc.DrawLine(rc.left, rc.bottom, rc.right, rc.bottom)
        dc.DrawLine(rc.left, rc.top, rc.left, rc.bottom)
        dc.DrawLine(rc.right - 1, rc.top, rc.right - 1, rc.bottom)
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DHILIGHT)))
        dc.DrawLine(rc.left, rc.top, rc.right, rc.top)
        dc.DrawLine(rc.left + 1, rc.top, rc.left + 1, rc.bottom)
        dc.DrawLine(rc.right, rc.top, rc.right, rc.bottom)

    def DrawBorder(self, dc, p):
        rc = p.GetRect()

        # top & bottom border
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DSHADOW)))
        dc.DrawLine(rc.left, rc.bottom, rc.right, rc.bottom)
        dc.DrawLine(rc.left, rc.top, rc.left, rc.bottom)
        dc.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DHILIGHT)))
        dc.DrawLine(rc.left, rc.top, rc.right, rc.top)
        dc.DrawLine(rc.left + 1, rc.top, rc.left + 1, rc.bottom)

        # draw selected box
        if p.activated:
            dc.SetPen(wx.Pen(wx.BLACK, 1, wx.PENSTYLE_DOT))
            dc.SetBrush(wx.Brush(wx.BLACK, wx.BRUSHSTYLE_TRANSPARENT))
            dc.DrawRectangle(p.GetRect())

    def DrawItem(self, dc, p):
        """draw the property"""
        if not p.IsVisible():
            return

        dc.SetBackgroundMode(wx.TRANSPARENT)

        self.PrepareDrawRect(p)

        self.DrawBackground(dc, p)

        self.DrawExpansion(dc, p)
        self.DrawLabel(dc, p)

        # separator does not have radio button, splitter bar and value sections
        if not p.IsSeparator():
            self.DrawSplitter(dc, p)
            self.DrawValue(dc, p)

        self.DrawBorder(dc, p)
